"""
default settings and a static container which loads and holds the settings from settings.json
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import os
import sys
import json
import logging
import imp  # TODO deprecated in Python 3

import colorama
from colorama import Fore

class SettingsException(Exception):
    pass


pyqt4_installed = False
try:
    imp.find_module("PyQt4")
    pyqt4_installed = True
except ImportError:
    pass
    

DEFAULT_SETTINGS_DICT = {
    "plot_xyz_realistic": True,
    "plot_backend": "Qt4Agg" if pyqt4_installed else "TkAgg",
    "plot_hideref": False,
    "plot_linewidth": 1.5,
    "plot_usetex": False,
    "plot_fontfamily": "sans-serif",
    "plot_fontsize": 12,
    "plot_split": False,
    "plot_figsize": [8, 8],
    "plot_info_text": False,
    "plot_trajectory_cmap": "jet",
    "plot_multi_cmap": "none",
    "plot_invert_xaxis": False,
    "plot_invert_yaxis": False,
    "plot_seaborn_style": "darkgrid",
    "plot_export_format": "pdf",
    "table_export_format": "csv",
    "table_export_transpose": True,
    "save_traj_in_zip": False,
    "logging_format": "%(message)s",
    "logfile_enabled": False
}

DEFAULT_SETTINGS_HELP = '''
plot_backend
    matplotlib backend - tabs are supported with TkAgg and Qt4Agg
plot_hideref
    hide reference trajectory in trajectory plots of metrics
plot_linewidth
    matplotlib supported line width value
plot_usetex
    use LaTeX renderer for plots
plot_fontfamily
    matplotlib supported font family string
plot_fontsize
    matplotlib supported font size integer
plot_split
    show / save each figure separately
    default: window with tabs from TkAgg backend
plot_figsize
    the default size of one (sub)plot figure (width, height)
plot_xyz_realistic
    equal axes aspect ratio for more realistic trajectory plots in xyz plot mode
plot_info_text
    allow text boxes with additional infos below the plots
plot_seaborn_style
    whitegrid, darkgrid, white, dark
plot_trajectory_cmap
    matplotlib color map for mapping values on a trajectory
plot_multi_cmap matplotlib
    color map for coloring plots from multiple data sources
    "none" will use default color cycle
plot_export_format
    matplotlib supported file format for exporting plots
table_export_format
    format for exporting tables (csv, excel, latex)
table_export_transpose
    transpose tables for export
save_traj_in_zip
    backup trajectories in result zip files (increases size)
logging_format
    format string for the logging module (console only)
logfile_enabled
    whether to write a logfile to the home folder
'''


class SettingsContainer(dict):
    def __init__(self, settings_path, data=None):
        super(SettingsContainer, self).__init__()
        try:
            if not data:
                with open(settings_path) as settings_file:
                    data = json.load(settings_file)
            for k, v in data.items():
                setattr(self, k, v)
        except Exception as e:
            logging.error(str(e))
            raise SettingsException("fatal: failed to load package settings file " + settings_path)

    def __getattr__(self, attr):
        # allow dot access
        return self[attr]

    def __setattr__(self, attr, value):
        # allow dot access
        self[attr] = value


def merge_dicts(first, second, soft=False):
    if soft:
        first.update({k: v for k, v in second.items() if not k in first})
    else:
        first.update(second)
    return first


PACKAGE_BASE_PATH = os.path.abspath(__file__ + "/../../")
PACKAGE_VERSION = open(os.path.join(PACKAGE_BASE_PATH, "version")).read()

USER_ASSETS_PATH = os.path.join(os.path.expanduser('~'), ".evo")
USER_ASSETS_VERSION_PATH = os.path.join(USER_ASSETS_PATH, "assets_version")

DEFAULT_PATH = os.path.join(USER_ASSETS_PATH, "settings.json")
DEFAULT_LOGFILE_PATH = os.path.join(USER_ASSETS_PATH, "evo.log")

# initialize .evo user folder after first installation (or if it was deleted)
if not os.path.isdir(USER_ASSETS_PATH):
    os.makedirs(USER_ASSETS_PATH)

if not os.path.exists(USER_ASSETS_VERSION_PATH):
    open(os.path.join(USER_ASSETS_PATH, "assets_version"), 'w').write(PACKAGE_VERSION)

if not os.path.exists(DEFAULT_PATH):
    try:
        with open(DEFAULT_PATH, 'w') as cfg_file:
            cfg_file.write(json.dumps(DEFAULT_SETTINGS_DICT, indent=4, sort_keys=True))
        print(Fore.LIGHTYELLOW_EX + "initialized new " + DEFAULT_PATH + Fore.RESET)
    except:
        print(Fore.LIGHTRED_EX
            + "fatal: failed to write package settings file " + DEFAULT_PATH + Fore.RESET)
        raise

if not open(USER_ASSETS_VERSION_PATH).read() == PACKAGE_VERSION:
    old_settings = json.loads(open(DEFAULT_PATH).read())
    updated_settings = merge_dicts(old_settings, DEFAULT_SETTINGS_DICT, soft=True)
    with open(DEFAULT_PATH, 'w') as cfg_file:
        cfg_file.write(json.dumps(updated_settings, indent=4, sort_keys=True))
    open(os.path.join(USER_ASSETS_PATH, "assets_version"), 'w').write(PACKAGE_VERSION)
    print(Fore.LIGHTYELLOW_EX + "updated outdated " + DEFAULT_PATH + Fore.RESET)


# load the user settings into this container
SETTINGS = SettingsContainer(DEFAULT_PATH)


class DefaultConsoleFormatter(logging.Formatter):
    def __init__(self, fmt="%(msg)s"):
        logging.Formatter.__init__(self, fmt)
        colorama.init()
        self.error_fmt  = Fore.LIGHTRED_EX + "[%(levelname)s]" + Fore.RESET + "\n%(msg)s"
        self.warning_fmt  = Fore.LIGHTYELLOW_EX + "[%(levelname)s]" + Fore.RESET + "\n%(msg)s"
        self.info_fmt = fmt
        self.debug_fmt  = fmt

    def format(self, record):
        if record.levelno == logging.ERROR:
            self._fmt = self.error_fmt
        elif record.levelno == logging.WARNING:
            self._fmt = self.warning_fmt
        elif record.levelno == logging.INFO:
            self._fmt = self.info_fmt
        elif record.levelno == logging.DEBUG:
            self._fmt = self.debug_fmt
        result = logging.Formatter.format(self, record)
        return result


def configure_logging(verbose=False, silent=False, debug=False, 
                      fmt=None, dbg_fmt=None, file_path=None, mark_entry=True):
    if fmt is None:
        fmt = SETTINGS.logging_format
    if dbg_fmt is None:
        dbg_fmt = "[%(levelname)s][%(asctime)s][%(module)s.%(funcName)s():%(lineno)s]\n%(message)s"
    # if enabled, always log in debug mode to logfile
    if not SETTINGS.logfile_enabled:
        logfile = os.devnull
    elif file_path is None:
        logfile = DEFAULT_LOGFILE_PATH
    else:
        logfile = file_path
    logging.basicConfig(level=logging.DEBUG, format=dbg_fmt, filename=logfile, filemode='a')
    version = open(os.path.join(PACKAGE_BASE_PATH, "version")).read().rstrip()
    if mark_entry:
        logging.debug("#"*80 + "\nlog entry - evo " + version + "\n" + "#"*80)
    # console output based on user's choice
    if debug or verbose and not silent:
        console_level = logging.DEBUG
    elif not silent:
        console_level = logging.INFO
    else: 
        console_level = logging.WARNING
    console_handler = logging.StreamHandler(stream=sys.stdout)
    console_formatter = logging.Formatter(dbg_fmt) if debug else DefaultConsoleFormatter(fmt)
    console_handler.setFormatter(console_formatter)
    console_handler.setLevel(console_level)
    # add console_handler to default root logger
    logging.getLogger('').addHandler(console_handler)
    # and a global console logger that can be retreived by its name
    console_logger = logging.getLogger("console")
    console_logger.addHandler(console_handler)
    console_logger.setLevel(console_level)
    if debug:
        import getpass as gp
        import platform as pf
        logging.debug("system info:"
                      "\nPython %s\n%s\n%s\n"
                      % (pf.python_version(), pf.platform(), gp.getuser() + "@" + pf.node()))



#!/usr/bin/env python
# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
main executable for viewing result files from the trajectory metric apps
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

SEP = "-" * 80  # separator line


def parser():
    import argparse
    basic_desc = "tool for processing one or multiple result files from ape or rpe"
    lic = "(c) michael.grupp@tum.de"
    main_parser = argparse.ArgumentParser(description="%s %s" % (basic_desc, lic))
    output_opts = main_parser.add_argument_group("output options")
    usability_opts = main_parser.add_argument_group("usability options")
    main_parser.add_argument("result_files", help="one or multiple result files", nargs='+')
    main_parser.add_argument("--use_abs_time",
                             help="use absolute timestamps instead of the seconds from start "
                                  "(if trajectories were saved to result file)",
                             action="store_true")
    output_opts.add_argument("-p", "--plot", help="show plot window", action="store_true")
    output_opts.add_argument("--plot_markers", help="plot with circle markers", action="store_true")
    output_opts.add_argument("--save_plot", help="path to save plot", default=None)
    output_opts.add_argument("--serialize_plot", help="path to serialize plot (experimental)",
                             default=None)
    output_opts.add_argument("--save_table",
                             help="path to a file to save the results in a table", default=None)
    usability_opts.add_argument("--no_warnings", help="no warnings requiring user confirmation",
                                action="store_true")
    usability_opts.add_argument("-v", "--verbose", help="verbose output", action="store_true")
    usability_opts.add_argument("--silent", help="don't print any output", action="store_true")
    usability_opts.add_argument("--debug", help="verbose output with additional debug info",
                                action="store_true")
    usability_opts.add_argument("-c", "--config",
                                help=".json file with parameters (priority over command line args)")
    return main_parser


def run(args):
    import os
    import sys
    import logging

    import pandas as pd
    import numpy as np
    from natsort import natsorted

    from evo.tools import file_interface, user, settings
    from evo.tools.settings import SETTINGS

    settings.configure_logging(args.verbose, args.silent, args.debug)
    if args.debug:
        import pprint
        logging.debug("main_parser config:\n"
                      + pprint.pformat({arg: getattr(args, arg) for arg in vars(args)}) + "\n")

    # store data in Pandas data frames for easier analysis
    raw_df = pd.DataFrame()
    stat_df = pd.DataFrame()
    info_df = pd.DataFrame()
    use_seconds = False
    for result_file in args.result_files:
        logging.debug(SEP)
        info_dict, stat_dict, err_array, sec_from_start, traj_ref, traj_est \
            = file_interface.load_res_file(result_file, True)
        short_est_name = os.path.splitext(os.path.basename(info_dict["est_name"]))[0]
        if not args.no_warnings and (short_est_name in info_df.columns):
            logging.warning("double entry detected: " + short_est_name)
            if not user.confirm("ignore? enter 'y' to go on or any other key to quit"):
                sys.exit()
        if SETTINGS.plot_usetex:
            short_est_name = short_est_name.replace("_", "\\_")
        if args.use_abs_time:
            if traj_est is None:
                msg = "no trajectory found in " + result_file \
                      + " but --use_abs_time requires the trajectory in the result file - " \
                      + "to let the metrics app include them run: evo_config set save_traj_in_zip"
                raise RuntimeError(msg)
            else:
                index = traj_est.timestamps
                use_seconds = True
        elif sec_from_start is not None:
            index = sec_from_start.tolist()
            use_seconds = True
        else:
            index = np.arange(0, err_array.shape[0])
        info_dict["traj. backup?"] = traj_ref is not None and traj_est is not None
        info_dict["res_file"] = result_file
        new_raw_df = pd.DataFrame({short_est_name: err_array.tolist()}, index=index)
        new_info_df = pd.DataFrame({short_est_name: info_dict})
        new_stat_df = pd.DataFrame({short_est_name: stat_dict})
        # natural sort num strings "10" "100" "20" -> "10" "20" "100"
        new_stat_df = new_stat_df.reindex(index=natsorted(new_stat_df.index))
        # column-wise concatenation
        raw_df = pd.concat([raw_df, new_raw_df], axis=1)
        info_df = pd.concat([info_df, new_info_df], axis=1)
        stat_df = pd.concat([stat_df, new_stat_df], axis=1)
        # if verbose: log infos of the current data
        logging.debug("\n" + info_df[short_est_name].ix["title"])
        show_fields = ["est_name", "ref_name", "label", "traj. backup?"]
        logging.debug(info_df[short_est_name].ix[show_fields].to_string() + "\n")


    logging.debug(SEP)
    logging.info("\nstatistics overview:\n" + stat_df.T.to_string(line_width=80) + "\n")

    # check titles
    first_title = info_df.ix["title", 0]
    first_res_file = info_df.ix["res_file", 0]
    if args.save_table or args.plot or args.save_plot:
        for short_est_name, column in info_df.iteritems():
            if column.ix["title"] != first_title and not args.no_warnings:
                logging.info(SEP)
                logging.warning("mismatching titles, you probably use data from different metrics")
                logging.warning("conflict:\n" + "<"*7 + " " + first_res_file + "\n" + first_title + "\n"
                                + "="*7 + "\n" + column.ix["title"] + "\n"
                                + ">"*7 + " " + column.ix["res_file"])
                logging.warning("only the first one will be used as the title!")
                if not user.confirm("plot/save anyway? - enter 'y' or any other key to exit"):
                    sys.exit()

    if args.save_table:
        logging.debug(SEP)
        if args.no_warnings or user.check_and_confirm_overwrite(args.save_table):
            table_fmt = SETTINGS.table_export_format
            if SETTINGS.table_export_transpose:
                getattr(stat_df.T, "to_" + table_fmt)(args.save_table)
            else:
                getattr(stat_df, "to_" + table_fmt)(args.save_table)
            logging.debug(table_fmt + " table saved to: " + args.save_table)

    if args.plot or args.save_plot or args.serialize_plot:
        # check if data has NaN "holes" due to different indices
        inconsistent = raw_df.isnull().values.any()
        if inconsistent and not args.no_warnings:
            logging.debug(SEP)
            logging.warning(
                "data lengths/indices are not consistent, plotting probably makes no sense")
            if not user.confirm("plot anyway? - enter 'y' or any other key to exit"):
                sys.exit()

        from evo.tools import plot
        import matplotlib.pyplot as plt
        import seaborn as sns
        import math
        from scipy import stats

        # use default plot settings
        figsize = (SETTINGS.plot_figsize[0], SETTINGS.plot_figsize[1])
        use_cmap = SETTINGS.plot_multi_cmap.lower() != "none"
        colormap = SETTINGS.plot_multi_cmap if use_cmap else None
        linestyles = ["-o" for x in args.result_files] if args.plot_markers else None

        # labels according to first dataset
        title = first_title
        if "xlabel" in info_df.ix[:, 0].index:
            index_label = info_df.ix["xlabel", 0]
        else:
            index_label = "$t$ (s)" if use_seconds else "index"
        metric_label = info_df.ix["label", 0]

        plot_collection = plot.PlotCollection(title)
        # raw value plot
        fig_raw = plt.figure(figsize=figsize)
        # handle NaNs from concat() above
        raw_df.interpolate(method="index").plot(ax=fig_raw.gca(), colormap=colormap,
                                                style=linestyles, title=first_title)
        plt.xlabel(index_label)
        plt.ylabel(metric_label)
        plt.legend(frameon=True)
        plot_collection.add_figure("raw", fig_raw)

        # statistics plot
        fig_stats = plt.figure(figsize=figsize)
        exclude = stat_df.index.isin(["sse"])  # don't plot sse
        stat_df[~exclude].plot(kind="barh", ax=fig_stats.gca(), colormap=colormap, stacked=False)
        plt.xlabel(metric_label)
        plt.legend(frameon=True)
        plot_collection.add_figure("stats", fig_stats)

        # grid of distribution plots
        raw_tidy = pd.melt(raw_df, value_vars=list(raw_df.columns.values), var_name="estimate",
                           value_name=metric_label)
        col_wrap = 2 if len(args.result_files) <= 2 else math.ceil(len(args.result_files) / 2.0)
        dist_grid = sns.FacetGrid(raw_tidy, col="estimate", col_wrap=col_wrap)
        dist_grid.map(sns.distplot, metric_label)  # fits=stats.gamma
        plot_collection.add_figure("histogram", dist_grid.fig)

        # box plot
        fig_box = plt.figure(figsize=figsize)
        ax = sns.boxplot(x=raw_tidy["estimate"], y=raw_tidy[metric_label], ax=fig_box.gca())
        # ax.set_xticklabels(labels=[item.get_text() for item in ax.get_xticklabels()], rotation=30)
        plot_collection.add_figure("box_plot", fig_box)

        # violin plot
        fig_violin = plt.figure(figsize=figsize)
        ax = sns.violinplot(x=raw_tidy["estimate"], y=raw_tidy[metric_label], ax=fig_violin.gca())
        # ax.set_xticklabels(labels=[item.get_text() for item in ax.get_xticklabels()], rotation=30)
        plot_collection.add_figure("violin_histogram", fig_violin)

        if args.plot:
            plot_collection.show()
        if args.save_plot:
            logging.debug(SEP)
            plot_collection.export(args.save_plot, confirm_overwrite=not args.no_warnings)
        if args.serialize_plot:
            logging.debug(SEP)
            plot_collection.serialize(args.serialize_plot, confirm_overwrite=not args.no_warnings)


if __name__ == '__main__':
    from evo import entry_points
    entry_points.res()



# -*- coding: UTF8 -*-
"""
low- and high-level read/write functions for different file formats
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import io
import zipfile
import csv
import logging
import os
import json

import numpy as np

from evo.algorithms import sync
from evo.algorithms import trajectory
from evo.algorithms.trajectory import PosePath3D, PoseTrajectory3D
from evo.tools import user
from evo.tools.settings import SETTINGS


class FileInterfaceException(Exception):
    pass


def csv_read_matrix(file_path, delim=',', comment_str="#"):
    """
    directly parse a csv-like file into a matrix
    :param file_path: path of csv file (or file handle)
    :param delim: delimiter character
    :param comment_str: string indicating a comment line to ignore
    :return: 2D list with raw data (string)
    """
    if hasattr(file_path, 'read'):  # if file handle
        generator = (line for line in file_path if not line.startswith(comment_str))
        reader = csv.reader(generator, delimiter=delim)
        mat = [row for row in reader]
    else:
        if not os.path.isfile(file_path):
            raise FileInterfaceException("csv file " + str(file_path) + " does not exist")
        with open(file_path) as f:
            generator = (line for line in f if not line.startswith(comment_str))
            reader = csv.reader(generator, delimiter=delim)
            mat = [row for row in reader]
    return mat


def read_tum_trajectory_file(file_path):
    """
    parses trajectory file in TUM format (timestamp tx ty tz qx qy qz qw)
    :param file_path: the trajectory file path (or file handle)
    :return: trajectory.PoseTrajectory3D object
    """
    mat = np.array(csv_read_matrix(file_path, delim=" ", comment_str="#")).astype(float)
    if mat.shape[1] != 8:
        raise FileInterfaceException("TUM trajectory files must have 8 entries per row")
    stamps = mat[:, 0]  # n x 1
    xyz = mat[:, 1:4]  # n x 3
    quat = mat[:, 4:]  # n x 4
    quat = np.roll(quat, 1, axis=1)  # shift 1 column -> w in front column
    if not hasattr(file_path, 'read'):  # if not file handle
        logging.debug("loaded " + str(len(stamps)) + " stamps and poses from: " + file_path)
    return PoseTrajectory3D(xyz, quat, stamps)


def write_tum_trajectory_file(file_path, traj, quat_is_wxyz=True, confirm_overwrite=False):
    """
    :param file_path: desired text file for trajectory (string or handle)
    :param traj: trajectory.PoseTrajectory3D
    :param quat_is_wxyz: set to False if you use xyzw quaternion order
    :param confirm_overwrite: whether to require user interaction to overwrite existing files
    """
    if isinstance(file_path, str) and confirm_overwrite:
        if not user.check_and_confirm_overwrite(file_path):
            return
    if not isinstance(traj, PoseTrajectory3D):
        raise FileInterfaceException("trajectory must be a PoseTrajectory3D object")
    stamps = traj.timestamps
    xyz = traj.positions_xyz
    quat = traj.orientations_quat_wxyz
    if quat_is_wxyz:
        quat = np.roll(quat, -1, axis=1)  # shift -1 column -> w in back column
    mat = np.column_stack((stamps, xyz, quat))
    np.savetxt(file_path, mat, delimiter=" ")
    if isinstance(file_path, str):
        logging.info("trajectory saved to: " + file_path)


def read_kitti_poses_file(file_path):
    """
    parses pose file in KITTI format (first 3 rows of SE(3) matrix per line)
    :param file_path: the trajectory file path (or file handle)
    :return: trajectory.PosePath3D
    """
    mat = np.array(csv_read_matrix(file_path, delim=" ", comment_str="#")).astype(float)
    if mat.shape[1] != 12:
        raise FileInterfaceException("KITTI pose files must have 12 entries per row")
    poses = [np.array([[r[0], r[1], r[2], r[3]],
                       [r[4], r[5], r[6], r[7]],
                       [r[8], r[9], r[10], r[11]],
                       [0, 0, 0, 1]]) for r in mat]
    if not hasattr(file_path, 'read'):  # if not file handle
        logging.debug("loaded " + str(len(poses)) + " poses from: " + file_path)
    return PosePath3D(poses_se3=poses)


def write_kitti_poses_file(file_path, traj, confirm_overwrite=False):
    """
    :param file_path: desired text file for trajectory (string or handle)
    :param traj: trajectory.PosePath3D or trajectory.PoseTrajectory3D
    :param confirm_overwrite: whether to require user interaction to overwrite existing files
    """
    if isinstance(file_path, str) and confirm_overwrite:
        if not user.check_and_confirm_overwrite(file_path):
            return
    poses_flat = [p.flatten()[:-4] for p in traj.poses_se3]  # first 3 rows flattened
    np.savetxt(file_path, poses_flat, delimiter=' ')
    if isinstance(file_path, str):
        logging.info("poses saved to: " + file_path)


def read_euroc_csv_trajectory(file_path):
    """
    parses ground truth trajectory from EuRoC MAV state estimate .csv
    :param file_path: <sequence>/mav0/state_groundtruth_estimate0/data.csv
    :return: trajectory.PoseTrajectory3D object
    """
    mat = np.array(csv_read_matrix(file_path, delim=",", comment_str="#")).astype(float)
    if mat.shape[1] != 17:
        raise FileInterfaceException("EuRoC MAV state ground truth must have 17 entries per row")
    stamps = np.divide(mat[:, 0], 1e9)  # n x 1  -  nanoseconds to seconds
    xyz = mat[:, 1:4]  # n x 3
    quat = mat[:, 4:8]  # n x 4
    logging.debug("loaded " + str(len(stamps)) + " stamps and poses from: " + file_path)
    return PoseTrajectory3D(xyz, quat, stamps)


def read_bag_trajectory(bag_handle, topic):
    """
    :param bag_handle: opened bag handle, from rosbag.Bag(...)
    :param topic: geometry_msgs/PoseStamped topic
    :return: trajectory.PoseTrajectory3D
    """
    if not bag_handle.get_message_count(topic) > 0:
        raise FileInterfaceException("no messages for topic '" + topic + "' in bag")
    stamps, xyz, quat = [], [], []
    for topic, msg, t in bag_handle.read_messages(topic):
        stamps.append(t.secs + (t.nsecs * 1e-9))
        xyz.append([msg.pose.position.x, msg.pose.position.y, msg.pose.position.z])
        quat.append([msg.pose.orientation.x, msg.pose.orientation.y, 
                     msg.pose.orientation.z, msg.pose.orientation.w])
    logging.debug("loaded " + str(len(stamps)) + " geometry_msgs/PoseStamped messages" 
                  + " of topic: " + topic)
    quat = np.roll(quat, 1, axis=1)  # shift 1 column -> w in front column
    return PoseTrajectory3D(xyz, quat, stamps)


def write_bag_trajectory(bag_handle, traj, topic_name):
    """
    :param bag_handle: opened bag handle, from rosbag.Bag(...)
    :param traj: trajectory.PoseTrajectory3D
    :param topic_name: the desired topic name for the trajectory
    """
    import rospy
    from geometry_msgs.msg import PoseStamped
    if not isinstance(traj, PoseTrajectory3D):
        raise FileInterfaceException("trajectory must be a PoseTrajectory3D object")
    for stamp, xyz, quat in zip(traj.timestamps, traj.positions_xyz, traj.orientations_quat_wxyz):
        p = PoseStamped()
        p.header.stamp = rospy.Time.from_sec(stamp)
        p.header.frame_id = topic_name.replace("/", "")
        p.pose.position.x = xyz[0]
        p.pose.position.y = xyz[1]
        p.pose.position.z = xyz[2]
        p.pose.orientation.w = quat[0]
        p.pose.orientation.x = quat[1]
        p.pose.orientation.y = quat[2]
        p.pose.orientation.z = quat[3]
        bag_handle.write(topic_name, p, t=p.header.stamp)
    logging.info("saved geometry_msgs/PoseStamped topic: " + topic_name)


def load_assoc_tum_trajectories(ref_file, est_file, max_diff=0.01, offset_2=0.0, invert=False):
    """
    parses two trajectory files in TUM format (timestamp tx ty tz qx qy qz qw)
    and returns the data with associated (matching) timestamps according to the time parameters
    :param ref_file: first trajectory file
    :param est_file: second trajectory file
    :param max_diff: max. allowed absolute time difference for associating
    :param offset_2: optional time offset of second timestamp vector
    :param invert: set to True to match from longer list to short (default: short to longer list)
    :return: trajectory.PoseTrajectory3D objects traj_ref and traj_est
    """
    traj_ref = read_tum_trajectory_file(ref_file)
    traj_est = read_tum_trajectory_file(est_file)
    return sync.associate_trajectories(traj_ref, traj_est, max_diff, offset_2,
                                       invert, ref_file, est_file)


def load_assoc_euroc_trajectories(ref_file, est_file, max_diff=0.01, offset_2=0.0, invert=False):
    """
    parses ground truth trajectory from EuRoC MAV state estimate .csv and estimated TUM trajectory
    and returns the data with associated (matching) timestamps according to the time parameters
    :param ref_file: ground truth: <sequence>/mav0/state_groundtruth_estimate0/data.csv
    :param est_file: estimated TUM trajectory file
    :param max_diff: max. allowed absolute time difference for associating
    :param offset_2: optional time offset of second timestamp vector
    :param invert: set to True to match from longer list to short (default: short to longer list)
    :return: trajectory.PoseTrajectory3D objects traj_ref and traj_est
    """
    traj_ref = read_euroc_csv_trajectory(ref_file)
    traj_est = read_tum_trajectory_file(est_file)
    return sync.associate_trajectories(traj_ref, traj_est, max_diff, offset_2, invert,
                                       ref_file, est_file)


def load_assoc_bag_trajectories(bag_handle, ref_topic, est_topic,
                                max_diff=0.01, offset_2=0.0, invert=False):
    """
    reads trajectory data from a ROS bag file with two geometry_msgs/PoseStamped topics
    and returns the data with associated (matching) timestamps according to the time parameters
    :param bag_handle: opened bag handle, from rosbag.Bag(...)
    :param ref_topic: first geometry_msgs/PoseStamped topic
    :param est_topic: second geometry_msgs/PoseStamped topic
    :param max_diff: max. allowed absolute time difference for associating
    :param offset_2: optional time offset of second timestamp vector
    :param invert: set to True to match from longer list to short (default: short to longer list)
    :return: trajectory.PoseTrajectory3D objects traj_ref and traj_est
    """
    traj_ref = read_bag_trajectory(bag_handle, ref_topic)
    traj_est = read_bag_trajectory(bag_handle, est_topic)
    return sync.associate_trajectories(traj_ref, traj_est, max_diff, offset_2, invert,
                                       ref_topic, est_topic)


def save_res_file(zip_path, pe_metric, pe_statistics, title, ref_name, est_name,
                  seconds_from_start=None, traj_ref=None, traj_est=None, xlabel=None,
                  confirm_overwrite=False):
    """
    save results of a pose error metric (pe_metric) to a zip file
    :param zip_path: path to zip file
    :param pe_metric: instance of metrics.APE or metrics.RPE
    :param pe_statistics: dictionary with statistics from pe_metric.get_[all_]statistic[s]
    :param title: the title of the experiment
    :param ref_name: name of the reference (ground_truth)
    :param est_name: name of the estimate
    :param seconds_from_start: optional time array (useful for plotting the time instead of index)
    :param traj_ref: optional - save trajectory that was used as the reference (requires traj_est)
    :param traj_est: optional - save trajectory that was used as the estimate (requires traj_ref)
    :param xlabel: optional custom (plot) xlabel to save
    :param confirm_overwrite: whether to require user interaction to overwrite existing files
    """
    from tempfile import TemporaryFile
    from evo.algorithms.metrics import APE
    logging.debug("saving results to " + zip_path + "...")
    if confirm_overwrite and not user.check_and_confirm_overwrite(zip_path):
        return
    with zipfile.ZipFile(zip_path, 'w') as archive:
        pe_type = "APE" if isinstance(pe_metric, APE) else "RPE"
        info = {"title": title, "ref_name": ref_name, "est_name": est_name,
                "label": pe_type + (" (" + pe_metric.unit.value + ")") if pe_metric.unit else ""}
        if xlabel:
            info["xlabel"] = xlabel
        archive.writestr("info.json", json.dumps(info))
        archive.writestr("stats.json", json.dumps(pe_statistics))
        # save np arrays in .npz format
        tmp_file_err = TemporaryFile()
        np.save(tmp_file_err, pe_metric.error)
        tmp_file_err.seek(0)
        archive.writestr("error_array.npz", tmp_file_err.read())
        tmp_file_err.close()
        if seconds_from_start:
            tmp_file_sec = TemporaryFile()
            if seconds_from_start:
                np.save(tmp_file_sec, seconds_from_start)
            tmp_file_sec.seek(0)
            archive.writestr("sec_from_start.npz", tmp_file_sec.read())
            tmp_file_sec.close()
        if SETTINGS.save_traj_in_zip and traj_ref is not None and traj_est is not None:
            tmp_file_ref = TemporaryFile()
            tmp_file_est = TemporaryFile()
            if type(traj_ref) is PosePath3D or type(traj_est) is PosePath3D:
                traj_type = ".kitti"
                write_kitti_poses_file(tmp_file_ref, traj_ref)
                write_kitti_poses_file(tmp_file_est, traj_est)
            else:
                traj_type = ".tum"
                write_tum_trajectory_file(tmp_file_ref, traj_ref)
                write_tum_trajectory_file(tmp_file_est, traj_est)
            tmp_file_ref.seek(0)
            tmp_file_est.seek(0)
            archive.writestr("traj_ref" + traj_type, tmp_file_ref.read())
            archive.writestr("traj_est" + traj_type, tmp_file_est.read())
            tmp_file_ref.close()
            tmp_file_est.close()


def load_res_file(zip_path, load_trajectories=False):
    """
    load contents of a result .zip file saved with save_res_file(...)
    :param zip_path: path to zip file
    :param load_trajectories: set to True to load also the (backup) trajectories
    :return: info (dictionary), statistics (dictionary),
             error_array (numpy array), seconds_from_start (list or None)
    """
    logging.debug("loading results from " + zip_path + "...")
    with zipfile.ZipFile(zip_path, mode='r') as archive:
        file_list = archive.namelist()
        if not {"error_array.npz", "info.json", "stats.json"} <= set(file_list):
            raise FileInterfaceException("incorrect zip file structure")
        seconds_from_start = None
        if "sec_from_start.npz" in file_list:
            with io.BytesIO(archive.read("sec_from_start.npz")) as f:
                seconds_from_start = np.load(f)
        traj_ref, traj_est = None, None
        if load_trajectories and "traj_ref.tum" and "traj_est.tum" in file_list:
            with io.TextIOWrapper(archive.open("traj_ref.tum", mode='r')) as f:
                traj_ref = read_tum_trajectory_file(f)
            with io.TextIOWrapper(archive.open("traj_est.tum", mode='r')) as f:
                traj_est = read_tum_trajectory_file(f)
        elif load_trajectories and "traj_ref.kitti" and "traj_est.kitti" in file_list:
            with io.TextIOWrapper(archive.open("traj_ref.kitti", mode='r')) as f:
                traj_ref = read_kitti_poses_file(f)
            with io.TextIOWrapper(archive.open("traj_est.kitti", mode='r')) as f:
                traj_est = read_kitti_poses_file(f)
        with io.BytesIO(archive.read("error_array.npz")) as f:
            error_array = np.load(f)
        info = json.loads(archive.read("info.json").decode("utf-8"))
        statistics = json.loads(archive.read("stats.json").decode("utf-8"))
        return info, statistics, error_array, seconds_from_start, traj_ref, traj_est


def load_transform_json(json_path):
    """
    load a transformation stored in xyz + quaternion format in a .json file
    :param json_path: path to the .json file
    :return: t (SE(3) matrix), xyz (position), quat (orientation quaternion)
    """
    import evo.algorithms.lie_algebra as lie
    import evo.algorithms.transformations as tr
    with open(json_path, 'r') as tf_file:
        data = json.load(tf_file)
        keys = ("x", "y", "z", "qx", "qy", "qz", "qw")
        if not all(key in data for key in keys):
            raise FileInterfaceException("invalid transform file - expected keys " + str(keys))
        xyz = np.array([data["x"], data["y"], data["z"]])
        quat = np.array([data["qw"], data["qx"], data["qy"], data["qz"]])
        t = lie.se3(lie.so3_from_se3(tr.quaternion_matrix(quat)), xyz)
        return t, xyz, quat



# -*- coding: UTF8 -*-
"""
algorithms for time synchronization
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy
import logging

import numpy as np

from evo.algorithms.trajectory import PoseTrajectory3D


class SyncException(Exception):
    pass


def matching_time_indices(stamps_1, stamps_2, max_diff=0.01, offset_2=0.0):
    """
    searches for the best matching timestamps of 2 lists and returns their list indices
    :param stamps_1: first vector of timestamps
    :param stamps_2: second vector of timestamps
    :param max_diff: max. allowed absolute time difference
    :param offset_2: optional offset of second vector
    :return: the indices of the matching stamps in stamps_1
    """
    matching_indices = []
    stamps_2 += offset_2
    for stamp in stamps_1:
        diffs = np.abs(stamps_2 - stamp)
        argmin = np.argmin(diffs)
        if diffs[argmin] <= max_diff:
            matching_indices.append(argmin)
    return matching_indices


def associate_trajectories(traj_1, traj_2, max_diff=0.01, offset_2=0.0, invert=False,
                           first_name="first trajectory", snd_name="estimated trajectory"):
    """
    synchronize two trajectories via matching timestamps (e.g. for TUM RGB-D dataset, ROS bags, ...)
    :param traj_1: trajectory.PoseTrajectory3D object of first trajectory
    :param traj_2: trajectory.PoseTrajectory3D object of second trajectory
    :param max_diff: max. allowed absolute time difference for associating
    :param offset_2: optional time offset of second timestamp vector
    :param invert: set to True to match from longer list to short (default: short to longer list)
    :param first_name: optional name of reference trajectory for verbose/debug logging
    :param snd_name: optional name of estimated trajectory for verbose/debug logging
    :return: synchronized traj_1 and traj_2 of same length
    """
    if not isinstance(traj_1, PoseTrajectory3D) or not isinstance(traj_2, PoseTrajectory3D):
        raise SyncException("trajectories must be PoseTrajectory3D objects")
    if invert:
        logging.debug("using inverse matching logic")
    traj_1 = copy.deepcopy(traj_1)
    traj_2 = copy.deepcopy(traj_2)
    snd_longer = len(traj_2.timestamps) > len(traj_1.timestamps)
    max_pairs = len(traj_2.timestamps) if not snd_longer and not invert else len(traj_1.timestamps)
    if invert:
        snd_longer = not snd_longer
    traj_long = traj_2 if snd_longer else traj_1
    traj_short = traj_1 if snd_longer else traj_2

    # associate the timestamps of the shorter trajectory to the longer one
    # select matching data from longer one
    matching_indices = matching_time_indices(traj_short.timestamps, traj_long.timestamps,
                                             max_diff, offset_2 if snd_longer else -offset_2)
    traj_long.reduce_to_ids(matching_indices)

    # reversely select matching data from shorter one
    # (if longer one is now smaller than shorter one)
    matching_indices = matching_time_indices(traj_long.timestamps, traj_short.timestamps,
                                             max_diff, -offset_2 if snd_longer else offset_2)
    traj_short.reduce_to_ids(matching_indices)

    traj_1 = traj_short if snd_longer else traj_long
    traj_2 = traj_long if snd_longer else traj_short
    if len(matching_indices) == 0:
        raise SyncException("found no matching timestamps between "
                            + first_name + " and " + snd_name + " with "
                            + "max. time diff.: " + str(max_diff)
                            + " (s) and time offset: " + str(offset_2) + " (s)")
    logging.debug("found " + str(len(matching_indices)) + " of max. " + str(max_pairs)
                  + " possible matching timestamps between..."
                  + "\n\t" + first_name
                  + "\nand: \t" + snd_name
                  + "\n...with max. time diff.: " + str(max_diff)
                  + " (s) and time offset: " + str(offset_2) + " (s)")

    return traj_1, traj_2



#!/usr/bin/env python
# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
main executable for calculating the sub-sequence-wise averaged relative pose error (RPE) metric
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import logging

SEP = "-" * 80  # separator line


def parser():
    import argparse
    basic_desc = "sub-sequence-wise averaged relative pose error (RPE) metric app"
    lic = "(c) michael.grupp@tum.de"
    shared_parser = argparse.ArgumentParser(add_help=False)
    algo_opts = shared_parser.add_argument_group("algorithm options")
    output_opts = shared_parser.add_argument_group("output options")
    usability_opts = shared_parser.add_argument_group("usability options")
    algo_opts.add_argument("-r", "--pose_relation",
                           help="pose relation on which the RPE is based", default="trans_part",
                           choices=["trans_part", "angle_deg", "angle_rad"])
    algo_opts.add_argument("-m", "--mode", help="the mode on which sequence averaging is based",
                           default="path", choices=["speed", "path", "angle", "angular_speed"])
    algo_opts.add_argument("-b", "--bins", help="bin values of the sub-sequences",
                           nargs='+', type=float)
    algo_opts.add_argument("-t", "--tols",
                           help="rel. tolerances to include sub-sequences in the bins",
                           nargs='+', type=float)
    algo_opts.add_argument("-a", "--align", help="alignment with Umeyama's method (no scale) - "
                                                 "only useful for plotting with RPE",
                           action="store_true")
    algo_opts.add_argument("-s", "--correct_scale", help="correct scale with Umeyama's method",
                           action="store_true")
    output_opts.add_argument("-p", "--plot", help="show plot window", action="store_true")
    output_opts.add_argument("--save_plot", help="path to save plot", default=None)
    output_opts.add_argument("--serialize_plot", help="path to serialize plot (experimental)",
                             default=None)
    output_opts.add_argument("--save_results", help=".zip file path to store results")
    usability_opts.add_argument("--no_warnings", help="no warnings requiring user confirmation",
                                action="store_true")
    usability_opts.add_argument("-v", "--verbose", help="verbose output", action="store_true")
    usability_opts.add_argument("--silent", help="don't print any output", action="store_true")
    usability_opts.add_argument("--debug", help="verbose output with additional debug info",
                                action="store_true")
    usability_opts.add_argument("-c", "--config",
                                help=".json file with parameters (priority over command line args)")

    main_parser = argparse.ArgumentParser(description="%s %s" % (basic_desc, lic))
    sub_parsers = main_parser.add_subparsers(dest="subcommand")
    sub_parsers.required = True
    kitti_parser = sub_parsers.add_parser("kitti", description="%s for KITTI pose files - %s"
                                                               % (basic_desc, lic),
                                          parents=[shared_parser])
    kitti_parser.add_argument("ref_file", help="reference pose file (ground truth)")
    kitti_parser.add_argument("est_file", help="estimated pose file")

    tum_parser = sub_parsers.add_parser("tum", description="%s for TUM trajectory files - %s"
                                                           % (basic_desc, lic),
                                        parents=[shared_parser])
    tum_parser.add_argument("ref_file", help="reference trajectory file")
    tum_parser.add_argument("est_file", help="estimated trajectory file")
    tum_parser.add_argument("--t_max_diff",
                            help="maximum timestamp difference for data association",
                            default=0.01, type=float)
    tum_parser.add_argument("--t_offset", help="constant timestamp offset for data association",
                            default=0.0, type=float)

    euroc_parser = sub_parsers.add_parser("euroc", description="%s for EuRoC MAV .csv's - %s"
                                                               % (basic_desc, lic),
                                          parents=[shared_parser])
    euroc_parser.add_argument("state_gt_csv",
                              help="ground truth: <seq>/mav0/state_groundtruth_estimate0/data.csv")
    euroc_parser.add_argument("est_file", help="estimated trajectory file in TUM format")
    euroc_parser.add_argument("--t_max_diff",
                              help="maximum timestamp difference for data association",
                              default=0.01, type=float)
    euroc_parser.add_argument("--t_offset", help="constant timestamp offset for data association",
                              default=0.0, type=float)

    bag_parser = sub_parsers.add_parser("bag",
                                        description="%s for ROS bag files - %s" % (basic_desc, lic),
                                        parents=[shared_parser])
    bag_parser.add_argument("bag", help="ROS bag file")
    bag_parser.add_argument("ref_topic", help="reference geometry_msgs/PoseStamped topic")
    bag_parser.add_argument("est_topic", help="estimated geometry_msgs/PoseStamped topic")
    bag_parser.add_argument("--t_max_diff",
                            help="maximum timestamp difference for data association",
                            default=0.01, type=float)
    bag_parser.add_argument("--t_offset", help="constant timestamp offset for data association",
                            default=0.0, type=float)
    return main_parser


def main_rpe_for_each(traj_ref, traj_est, pose_relation, mode, bins, rel_tols,
                      align=False, correct_scale=False, ref_name="", est_name="",
                      show_plot=False, save_plot=None, save_results=None, no_warnings=False,
                      serialize_plot=None):

    from evo.algorithms import metrics
    from evo.algorithms import filters
    from evo.algorithms import trajectory
    from evo.tools import file_interface
    from evo.tools.settings import SETTINGS

    if not bins or not rel_tols:
        raise RuntimeError("bins and tolerances must have more than one element")
    if len(bins) != len(rel_tols):
        raise RuntimeError("bins and tolerances must have the same number of elements")
    if mode in {"speed", "angular_speed"} and traj_est is trajectory.PosePath3D:
        raise RuntimeError("timestamps are required for mode: " + mode)

    bin_unit = None
    if mode == "speed":
        bin_unit = metrics.VelUnit.meters_per_sec
    elif mode == "path":
        bin_unit = metrics.Unit.meters
    elif mode == "angle":
        bin_unit = metrics.Unit.degrees
    elif mode == "angular_speed":
        bin_unit = metrics.VelUnit.degrees_per_sec

    rpe_unit = None
    if pose_relation is metrics.PoseRelation.translation_part:
        rpe_unit = metrics.Unit.meters
    elif pose_relation is metrics.PoseRelation.rotation_angle_deg:
        rpe_unit = metrics.Unit.degrees
    elif pose_relation is metrics.PoseRelation.rotation_angle_rad:
        rpe_unit = metrics.Unit.radians

    correct_only_scale = correct_scale and not align
    if align or correct_scale:
        logging.debug(SEP)
        if correct_only_scale:
            logging.debug("correcting scale...")
        else:
            logging.debug("aligning using Umeyama's method..."
                          + (" (with scale correction)" if correct_scale else ""))
        traj_est = trajectory.align_trajectory(traj_est, traj_ref, correct_scale,
                                               correct_only_scale)

    results = []
    for bin, rel_tol, in zip(bins, rel_tols):
        logging.debug(SEP)
        logging.info(
            "calculating RPE for each sub-sequence of " + str(bin) + " (" + bin_unit.value + ")")

        tol = bin * rel_tol
        id_pairs = []
        if mode == "path":
            id_pairs = filters.filter_pairs_by_path(traj_ref.poses_se3, bin, tol, all_pairs=True)
        elif mode == "angle":
            id_pairs = filters.filter_pairs_by_angle(traj_ref.poses_se3, bin, tol, degrees=True)
        elif mode == "speed":
            id_pairs = filters.filter_pairs_by_speed(traj_ref.poses_se3, traj_ref.timestamps,
                                                     bin, tol)
        elif mode == "angular_speed":
            id_pairs = filters.filter_pairs_by_angular_speed(traj_ref.poses_se3,
                                                             traj_ref.timestamps, bin, tol, True)

        if len(id_pairs) == 0:
            raise RuntimeError("bin " + str(bin) + " (" + str(bin_unit.value) + ") "
                               + "produced empty index list - try other values")

        # calculate RPE with all IDs (delta 1 frames)
        data = (traj_ref, traj_est)
        # the delta here has nothing to do with the bin - 1f delta just to use all poses of the bin
        rpe_metric = metrics.RPE(pose_relation, delta=1, delta_unit=metrics.Unit.frames,
                                 all_pairs=True)
        rpe_metric.process_data(data, id_pairs)
        mean = rpe_metric.get_statistic(metrics.StatisticsType.mean)
        results.append(mean)

    if SETTINGS.plot_usetex:
        mode.replace("_", "\_")
    title = "mean RPE w.r.t. " + pose_relation.value + "\nfor different " + mode + " sub-sequences"
    if align and not correct_scale:
        title += "\n(with SE(3) Umeyama alignment)"
    elif align and correct_scale:
        title += "\n(with Sim(3) Umeyama alignment)"
    elif correct_only_scale:
        title += "\n(scale corrected)"
    else:
        title += "\n(not aligned)"
    logging.debug(SEP)
    logging.info("\n" + title + "\n")
    res_str = ""
    for bin, result in zip(bins, results):
        res_str += "{:>10}".format(str(bin) + "(" + bin_unit.value + ")")
        res_str += "\t" + "{0:.6f}".format(result) + "\n"
    logging.info(res_str)

    if show_plot or save_plot or serialize_plot:
        from evo.tools import plot
        import matplotlib.pyplot as plt
        plot_collection = plot.PlotCollection(title)
        fig = plt.figure(figsize=(SETTINGS.plot_figsize[0], SETTINGS.plot_figsize[1]))
        plot.error_array(fig, results, x_array=bins,
                         name="mean RPE" + (" (" + rpe_unit.value + ")") if rpe_unit else "",
                         marker="o", title=title,
                         xlabel=mode + " sub-sequences " + " (" + bin_unit.value + ")")
        # info text
        if SETTINGS.plot_info_text and est_name and ref_name:
            ax = fig.gca()
            ax.text(0, -0.12, "estimate:  " + est_name + "\nreference: " + ref_name,
                    transform=ax.transAxes, fontsize=8, color="gray")
        plt.title(title)
        plot_collection.add_figure("raw", fig)
        if show_plot:
            plot_collection.show()
        if save_plot:
            plot_collection.export(save_plot, confirm_overwrite=not no_warnings)
        if serialize_plot:
            logging.debug(SEP)
            plot_collection.serialize(serialize_plot, confirm_overwrite=not no_warnings)

    rpe_statistics = {bin: result for bin, result in zip(bins, results)}
    if save_results:
        logging.debug(SEP)

        # utility class to trick save_res_file
        class Metric:
            unit = rpe_unit
            error = results

        file_interface.save_res_file(save_results, Metric, rpe_statistics,
                                     title, ref_name, est_name, bins, traj_ref, traj_est,
                                     xlabel=mode + " sub-sequences " + " (" + bin_unit.value + ")",
                                     confirm_overwrite=not no_warnings)

    return rpe_statistics, results


def run(args):
    import sys
    from evo.algorithms import metrics
    from evo.tools import file_interface, settings

    # manually check bins and tols arguments to allow them to be in config files
    if not args.bins or not args.tols:
        logging.error("the following arguments are required: -b/--bins, -t/--tols")
        sys.exit(1)

    settings.configure_logging(args.verbose, args.silent, args.debug)
    if args.debug:
        import pprint
        logging.debug("main_parser config:\n"
                      + pprint.pformat({arg: getattr(args, arg) for arg in vars(args)}) + "\n")
    logging.debug(SEP)

    pose_relation = None
    if args.pose_relation == "trans_part":
        pose_relation = metrics.PoseRelation.translation_part
    elif args.pose_relation == "angle_deg":
        pose_relation = metrics.PoseRelation.rotation_angle_deg
    elif args.pose_relation == "angle_rad":
        pose_relation = metrics.PoseRelation.rotation_angle_rad

    traj_ref, traj_est, stamps_est = None, None, None
    ref_name, est_name = "", ""
    if args.subcommand == "tum":
        traj_ref, traj_est = file_interface.load_assoc_tum_trajectories(
            args.ref_file,
            args.est_file,
            args.t_max_diff,
            args.t_offset,
        )
        ref_name, est_name = args.ref_file, args.est_file
    elif args.subcommand == "kitti":
        traj_ref = file_interface.read_kitti_poses_file(args.ref_file)
        traj_est = file_interface.read_kitti_poses_file(args.est_file)
        ref_name, est_name = args.ref_file, args.est_file
    elif args.subcommand == "euroc":
        args.align = True
        logging.info("forcing trajectory alignment implicitly (EuRoC ground truth is in IMU frame)")
        logging.debug(SEP)
        traj_ref, traj_est = file_interface.load_assoc_euroc_trajectories(
            args.state_gt_csv,
            args.est_file,
            args.t_max_diff,
            args.t_offset,
        )
        ref_name, est_name = args.state_gt_csv, args.est_file
    elif args.subcommand == "bag":
        import rosbag
        logging.debug("opening bag file " + args.bag)
        bag = rosbag.Bag(args.bag, 'r')
        try:
            traj_ref, traj_est = file_interface.load_assoc_bag_trajectories(
                bag,
                args.ref_topic,
                args.est_topic,
                args.t_max_diff,
                args.t_offset,
            )
        finally:
            bag.close()
        ref_name, est_name = args.ref_topic, args.est_topic

    main_rpe_for_each(traj_ref, traj_est, pose_relation, args.mode, args.bins, args.tols,
                      args.align, args.correct_scale, ref_name, est_name,
                      args.plot, args.save_plot, args.save_results, args.no_warnings,
                      serialize_plot=args.serialize_plot)


if __name__ == '__main__':
    from evo import entry_points
    entry_points.rpe_for_each()



#!/usr/bin/env python
# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
main executable for trajectory analysis
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

SEP = "-" * 80


def parser():
    import argparse
    basic_desc = "trajectory analysis and manipulation tool"
    lic = "(c) michael.grupp@tum.de"
    shared_parser = argparse.ArgumentParser(add_help=False)
    algo_opts = shared_parser.add_argument_group("algorithm options")
    output_opts = shared_parser.add_argument_group("output options")
    usability_opts = shared_parser.add_argument_group("usability options")
    shared_parser.add_argument("-f", "--full_check", help="run all checks and print all stats",
                               action="store_true")
    algo_opts.add_argument("-a", "--align", help="alignment with Umeyama's method (no scale)"
                                                 " - requires --ref", action="store_true")
    algo_opts.add_argument("-s", "--correct_scale", help="scale correction with Umeyama's method"
                                                         " - requires --ref", action="store_true")
    algo_opts.add_argument("--n_to_align", help="the number of poses to use for Umeyama alignment, "
                                                "counted from the start (default: all)",
                           default=-1, type=int)
    algo_opts.add_argument("--transform_left",
                           help="path to a .json file with a transformation"
                                " to apply to the trajectories (left multiplicative)")
    algo_opts.add_argument("--transform_right",
                           help="path to a .json file with a transformation"
                                " to apply to the trajectories (right_multiplicative)")
    algo_opts.add_argument("--invert_transform", help="invert the transformation of the .json file",
                           action="store_true")
    algo_opts.add_argument("--ref", help="trajectory that will be marked/used as the reference")
    algo_opts.add_argument("--t_offset", help="add a constant timestamp offset",
                           default=0.0, type=float)
    output_opts.add_argument("-p", "--plot", help="show plot window", action="store_true")
    output_opts.add_argument("--plot_mode", help="the axes for  plot projection",
                             default=None, choices=["xy", "yx", "xz", "zx", "yz", "xyz"])
    output_opts.add_argument("--save_plot", help="path to save plot", default=None)
    output_opts.add_argument("--serialize_plot", help="path to serialize plot (experimental)",
                             default=None)
    output_opts.add_argument("--save_as_tum", help="save trajectories in TUM format (as *.tum)",
                             action="store_true")
    output_opts.add_argument("--save_as_kitti", help="save poses in KITTI format (as *.kitti)",
                             action="store_true")
    output_opts.add_argument("--save_as_bag", help="save trajectories in ROS bag as <date>.bag",
                             action="store_true")
    usability_opts.add_argument("--no_warnings", help="no warnings requiring user confirmation",
                                action="store_true")
    usability_opts.add_argument("-v", "--verbose", help="verbose output", action="store_true")
    usability_opts.add_argument("--silent", help="don't print any output", action="store_true")
    usability_opts.add_argument("--debug", help="verbose output with additional debug info",
                                action="store_true")
    usability_opts.add_argument("-c", "--config",
                                help=".json file with parameters (priority over command line args)")

    main_parser = argparse.ArgumentParser(description="%s %s" % (basic_desc, lic))
    sub_parsers = main_parser.add_subparsers(dest="subcommand")
    sub_parsers.required = True
    kitti_parser = sub_parsers.add_parser("kitti", description="%s for KITTI pose files - %s"
                                                               % (basic_desc, lic),
                                          parents=[shared_parser])
    kitti_parser.add_argument("pose_files", help="one or multiple pose files", nargs='+')

    tum_parser = sub_parsers.add_parser("tum", description="%s for TUM trajectory files - %s"
                                                           % (basic_desc, lic),
                                        parents=[shared_parser])
    tum_parser.add_argument("traj_files", help="one or multiple trajectory files", nargs='+')

    euroc_parser = sub_parsers.add_parser("euroc", description="%s for EuRoC MAV .csv's - %s"
                                                               % (basic_desc, lic),
                                          parents=[shared_parser])
    euroc_parser.add_argument("state_gt_csv",
                              help="<sequence>/mav0/state_groundtruth_estimate0/data.csv",
                              nargs='+')

    bag_parser = sub_parsers.add_parser("bag", description="%s for ROS bag files - %s"
                                                           % (basic_desc, lic),
                                        parents=[shared_parser])
    bag_parser.add_argument("bag", help="ROS bag file")
    bag_parser.add_argument("topics", help="multiple geometry_msgs/PoseStamped topics", nargs='*')
    bag_parser.add_argument("--all_topics", help="use all topics geometry_msgs/PoseStamped in bag",
                            action="store_true")
    return main_parser


# TODO refactor
def print_traj_info(name, traj, verbose=False, full_check=False):
    import os
    import logging
    from evo.algorithms import trajectory

    logging.info(SEP)
    logging.info("name:\t" + os.path.splitext(os.path.basename(name))[0])

    if verbose or full_check:
        infos = traj.get_infos()
        info_str = ""
        for info, value in sorted(infos.items()):
            info_str += "\n\t" + info + "\t" + str(value)
        logging.info("infos:" + info_str)
        if full_check:
            passed, details = traj.check()
            check_str = ""
            for test, result in sorted(details.items()):
                check_str += "\n\t" + test + "\t" + result
            logging.info("checks:" + check_str)
            stat_str = ""
            try:
                stats = traj.get_statistics()
                for stat, value in sorted(stats.items()):
                    if isinstance(value, float):
                        stat_str += "\n\t" + stat + "\t" + "{0:.6f}".format(value)
                    else:
                        stat_str += value
            except trajectory.TrajectoryException as e:
                stat_str += "\n\terror - " + str(e)
            logging.info("stats:" + stat_str)
    else:
        logging.info("infos:\t" + str(traj))


def run(args):
    import os
    import sys
    import logging
    import evo.algorithms.lie_algebra as lie
    from evo.algorithms import trajectory
    from evo.tools import file_interface, settings
    from evo.tools.settings import SETTINGS

    settings.configure_logging(verbose=args.verbose, silent=args.silent, debug=args.debug)
    if args.debug:
        import pprint
        logging.debug("main_parser config:\n"
                      + pprint.pformat({arg: getattr(args, arg) for arg in vars(args)}) + "\n")
    logging.debug(SEP)

    trajectories = []
    ref_traj = None
    if args.subcommand == "tum":
        for traj_file in args.traj_files:
            if traj_file != args.ref:
                trajectories.append((traj_file, file_interface.read_tum_trajectory_file(traj_file)))
        if args.ref:
            ref_traj = file_interface.read_tum_trajectory_file(args.ref)
    elif args.subcommand == "kitti":
        for pose_file in args.pose_files:
            if pose_file != args.ref:
                trajectories.append((pose_file, file_interface.read_kitti_poses_file(pose_file)))
        if args.ref:
            ref_traj = file_interface.read_kitti_poses_file(args.ref)
    elif args.subcommand == "euroc":
        for csv_file in args.state_gt_csv:
            if csv_file != args.ref:
                trajectories.append((csv_file, file_interface.read_euroc_csv_trajectory(csv_file)))
        if args.ref:
            ref_traj = file_interface.read_euroc_csv_trajectory(args.ref)
    elif args.subcommand == "bag":
        import rosbag
        bag = rosbag.Bag(args.bag)
        try:
            if args.all_topics:
                topic_info = bag.get_type_and_topic_info()
                topics = sorted([t for t in topic_info[1].keys() if topic_info[1][t][0] 
                                 == "geometry_msgs/PoseStamped" and t != args.ref])
                if len(topics) == 0:
                    logging.error("no geometry_msgs/PoseStamped topics found!")
                    sys.exit(1)
            else:
                topics = args.topics
                if not topics:
                    logging.warning("no topics used - specify topics or use the --all_topics flag")
                    sys.exit(1)
            for topic in topics:
                trajectories.append((topic, file_interface.read_bag_trajectory(bag, topic)))
            if args.ref:
                ref_traj = file_interface.read_bag_trajectory(bag, args.ref)
        finally:
            bag.close()
    else:
        raise RuntimeError("unsupported subcommand: " + args.subcommand)

    if args.transform_left or args.transform_right:
        tf_type = "left" if args.transform_left else "right"
        tf_path = args.transform_left if args.transform_left else args.transform_right
        t, xyz, quat = file_interface.load_transform_json(tf_path)
        logging.debug(SEP)
        if not lie.is_se3(t):
            logging.warning("not a valid SE(3) transformation!")
        if args.invert_transform:
            t = lie.se3_inverse(t)
        logging.debug("applying a " + tf_type + "-multiplicative transformation:\n" + str(t))
        for name, traj in trajectories:
            traj.transform(t, right_mul=args.transform_right)

    if args.align or args.correct_scale:
        if args.ref:
            if args.subcommand == "kitti":
                traj_tmp, ref_traj_tmp = trajectories, [ref_traj for n, t in trajectories]
            else:
                traj_tmp, ref_traj_tmp = [], []
                from evo.algorithms import sync
                for name, traj in trajectories:
                    logging.debug(SEP)
                    ref_assoc, traj_assoc = sync.associate_trajectories(ref_traj, traj,
                                                                        first_name="ref",
                                                                        snd_name=name)
                    ref_traj_tmp.append(ref_assoc)
                    traj_tmp.append((name, traj_assoc))
                    trajectories = traj_tmp
            correct_only_scale = args.correct_scale and not args.align
            trajectories_new = []
            for nt, ref_assoc in zip(trajectories, ref_traj_tmp):
                logging.debug(SEP)
                logging.debug("aligning " + nt[0] + " to " + args.ref + "...")
                trajectories_new.append((nt[0], trajectory.align_trajectory(nt[1], ref_assoc, 
                                        args.correct_scale, correct_only_scale, args.n_to_align)))
            trajectories = trajectories_new

    for name, traj in trajectories:
        if args.t_offset and traj.timestamps.shape[0] != 0:
            logging.debug(SEP)
            logging.info("adding time offset to " + name + ": " + str(args.t_offset) + " (s)")
            traj.timestamps += args.t_offset
        print_traj_info(name, traj, args.verbose, args.full_check)
    if (args.align or args.correct_scale) and not args.ref:
        logging.debug(SEP)
        logging.warning("can't align without a reference! (--ref)  *grunt*")
    if args.ref:
        print_traj_info(args.ref, ref_traj, args.verbose, args.full_check)

    if args.plot or args.save_plot or args.serialize_plot:
        from evo.tools.plot import PlotMode
        plot_mode = PlotMode.xyz if not args.plot_mode else PlotMode[args.plot_mode]
        import numpy as np
        from evo.tools import plot
        import matplotlib.pyplot as plt
        import matplotlib.cm as cm
        plot_collection = plot.PlotCollection("evo_traj - trajectory plot")
        fig_xyz, axarr_xyz = plt.subplots(3, sharex="col", figsize=tuple(SETTINGS.plot_figsize))
        fig_traj = plt.figure(figsize=tuple(SETTINGS.plot_figsize))
        if (args.align or args.correct_scale) and not args.ref:
            plt.xkcd(scale=2, randomness=4)
            fig_traj.suptitle("what if --ref?")
            fig_xyz.suptitle("what if --ref?")
        ax_traj = plot.prepare_axis(fig_traj, plot_mode)
        if args.ref:
            short_traj_name = os.path.splitext(os.path.basename(args.ref))[0]
            if SETTINGS.plot_usetex:
                short_traj_name = short_traj_name.replace("_", "\\_")
            plot.traj(ax_traj, plot_mode, ref_traj, '--', 'grey', short_traj_name,
                      alpha=0 if SETTINGS.plot_hideref else 1)
            plot.traj_xyz(axarr_xyz, ref_traj, '--', 'grey', short_traj_name,
                          alpha=0 if SETTINGS.plot_hideref else 1)
        cmap_colors = None
        if SETTINGS.plot_multi_cmap.lower() != "none":
            cmap = getattr(cm, SETTINGS.plot_multi_cmap)
            cmap_colors = iter(cmap(np.linspace(0, 1, len(trajectories))))
        for name, traj in trajectories:
            if cmap_colors is None:
                color = next(ax_traj._get_lines.prop_cycler)['color']
            else:
                color = next(cmap_colors)
            short_traj_name = os.path.splitext(os.path.basename(name))[0]
            if SETTINGS.plot_usetex:
                short_traj_name = short_traj_name.replace("_", "\\_")
            plot.traj(ax_traj, plot_mode, traj, '-', color, short_traj_name)
            if args.ref and isinstance(ref_traj, trajectory.PoseTrajectory3D):
                start_time = ref_traj.timestamps[0]
            else:
                start_time = None
            plot.traj_xyz(axarr_xyz, traj, '-', color, short_traj_name, start_timestamp=start_time)
        plt.tight_layout()
        plot_collection.add_figure("trajectories", fig_traj)
        plot_collection.add_figure("xyz_view", fig_xyz)
        if args.plot:
            plot_collection.show()
        if args.save_plot:
            logging.info(SEP)
            plot_collection.export(args.save_plot, confirm_overwrite=not args.no_warnings)
        if args.serialize_plot:
            logging.info(SEP)
            plot_collection.serialize(args.serialize_plot, confirm_overwrite=not args.no_warnings)

    if args.save_as_tum:
        logging.info(SEP)
        for name, traj in trajectories:
            dest = os.path.splitext(os.path.basename(name))[0] + ".tum"
            file_interface.write_tum_trajectory_file(dest, traj,
                                                     confirm_overwrite=not args.no_warnings)
        if args.ref:
            dest = os.path.splitext(os.path.basename(args.ref))[0] + ".tum"
            file_interface.write_tum_trajectory_file(dest, ref_traj,
                                                     confirm_overwrite=not args.no_warnings)
    if args.save_as_kitti:
        logging.info(SEP)
        for name, traj in trajectories:
            dest = os.path.splitext(os.path.basename(name))[0] + ".kitti"
            file_interface.write_kitti_poses_file(dest, traj,
                                                  confirm_overwrite=not args.no_warnings)
        if args.ref:
            dest = os.path.splitext(os.path.basename(args.ref))[0] + ".kitti"
            file_interface.write_kitti_poses_file(dest, ref_traj,
                                                  confirm_overwrite=not args.no_warnings)
    if args.save_as_bag:
        logging.info(SEP)
        import datetime
        import rosbag
        dest_bag_path = str(datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S.%f')) + ".bag"
        logging.info("saving trajectories to " + dest_bag_path + "...")
        bag = rosbag.Bag(dest_bag_path, 'w')
        try:
            for name, traj in trajectories:
                dest_topic = os.path.splitext(os.path.basename(name))[0]
                file_interface.write_bag_trajectory(bag, traj, dest_topic)
            if args.ref:
                dest_topic = os.path.splitext(os.path.basename(args.ref))[0]
                file_interface.write_bag_trajectory(bag, ref_traj, dest_topic)
        finally:
            bag.close()


if __name__ == '__main__':
    from evo import entry_points
    entry_points.traj()



#!/usr/bin/env python
# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
simple JSON configuration generator script for executables
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import os
import sys
import json
import argparse
import logging

import colorama
from pygments import highlight, lexers, formatters

from evo.tools import user, settings

SEP = "-" * 80


class ConfigError(Exception):
    pass


def reset_pkg_settings(pkg_config=settings.DEFAULT_PATH):
    with open(pkg_config, 'w') as cfg_file:
        print(pkg_config)
        cfg_file.write(json.dumps(settings.DEFAULT_SETTINGS_DICT, indent=4, sort_keys=True))


def log_info_dict_json(data_str, colored=True):
    data_str = json.dumps(data_str, indent=4, sort_keys=True)
    if colored and os.name != "nt":
        data_str = highlight(data_str, lexers.JsonLexer(), 
                             formatters.Terminal256Formatter(style="monokai"))
    logging.info(data_str)


def show(cfg_path, colored=True):
    with open(cfg_path) as cfg_file:
        log_info_dict_json(json.load(cfg_file), colored)


def merge_json_union(first, second, soft=False):
    with open(first, 'r+') as f_1:
        cfg_1 = json.loads(f_1.read())
        with open(second) as f_2:
            cfg_2 = json.loads(f_2.read())
            cfg_1 = settings.merge_dicts(first, second, soft)
        f_1.truncate(0)
        f_1.seek(0)
        f_1.write(json.dumps(cfg_1, indent=4, sort_keys=True))


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def set_cfg(cfg_path, arg_list):
    with open(cfg_path) as cfg_file:
        data = json.load(cfg_file)
    max_idx = len(arg_list) - 1
    for i, arg in enumerate(arg_list):
        if arg in data:
            if i + 1 <= max_idx:
                if arg_list[i+1].lower() == "true":
                    data[arg] = True
                elif arg_list[i+1].lower() == "false":
                    data[arg] = False
                else:
                    values = []
                    for j in range(i + 1, max_idx + 1):
                        value = arg_list[j]
                        if value in data:
                            break
                        if is_number(value):
                            if int(float(value)) - float(value) != 0:
                                values.append(float(value))
                            else:
                                values.append(int(float(value)))
                        else:
                            values.append(value)
                    if len(values) == 1:
                        values = values[0]
                    data[arg] = not data[arg] if isinstance(data[arg], bool) else values
            elif i + 1 > max_idx or arg_list[i+1] in data:
                # toggle boolean parameter
                data[arg] = not data[arg] if isinstance(data[arg], bool) else data[arg]
    with open(cfg_path, 'w') as cfg_file:
        cfg_file.write(json.dumps(data, indent=4, sort_keys=True))


def generate(arg_list):
    data = {}
    max_idx = len(arg_list) - 1
    for i, arg in enumerate(arg_list):
        if arg.startswith("-"):
            arg = arg[1:] if not arg.startswith("--") else arg[2:]
            if (i + 1 <= max_idx and arg_list[i + 1].startswith("-")) or i + 1 > max_idx:
                data[arg] = True  # just a boolean flag
            else:
                values = []
                for j in range(i + 1, max_idx + 1):
                    value = arg_list[j]
                    if value.startswith("-"):
                        break
                    values.append(float(value) if is_number(value) else value)
                if len(values) == 1:
                    values = values[0]
                data[arg] = values
    return data


SET_HELP = '''
set parameters
Unless -c / --config is specified, the package settings will be used.

--Example--
If your configuration looks like this (via 'evo_config show'):

    {
        "plot_export_format": "svg"
        "plot_info_text": true,
    }
    
running:
    evo_config set plot_export_format png plot_info_text
will set it to:

    {
        "plot_export_format": "png"
        "plot_info_text": false,
    }
'''

GENERATE_HELP = '''
generate configuration files from command-line args

The configuration files are intended to hold command line
parameters used by the respective executables, e.g. 'evo_ape'.

--Example--
Running:
    evo_config generate --align --plot --plot_mode xz --verbose
will convert the arguments into the JSON format:
    {
        "align": true,
        "plot": true,
        "plot_mode": "xz",
        "verbose": true
    }
List arguments (--arg 1 2 3) are also supported.
To save the configuration, specify -o / --output.
'''


def main():
    import argcomplete
    basic_desc = "crappy configuration tool"
    lic = "(c) michael.grupp@tum.de"
    shared_parser = argparse.ArgumentParser(add_help=False)
    shared_parser.add_argument("--no_color", help="don't color output", 
                               action="store_false", default=True)
    main_parser = argparse.ArgumentParser(description="%s %s" % (basic_desc, lic))
    sub_parsers = main_parser.add_subparsers(dest="subcommand")
    sub_parsers.required = True

    show_parser = sub_parsers.add_parser("show", description="show configuration - %s" % lic,
                                         parents=[shared_parser])
    show_parser.add_argument("config",
                             help="optional config file to display (default: package settings)",
                             nargs='?')
    show_parser.add_argument("--brief", help="show only the .json data",
                             action="store_true")

    set_parser = sub_parsers.add_parser("set", description=SET_HELP, parents=[shared_parser],
                                        formatter_class=argparse.RawTextHelpFormatter)
    set_parser.add_argument("-c", "--config",
                            help="optional config file (default: package settings)", default=None)
    set_parser.add_argument("-m", "--merge",
                            help="optional config file (priority) for union merging", default=None)

    gen_parser = sub_parsers.add_parser("generate", description=GENERATE_HELP,
                                        parents=[shared_parser],
                                        formatter_class=argparse.RawTextHelpFormatter)
    gen_parser.add_argument("-o", "--out", help="path for config file to generate")

    reset_parser = sub_parsers.add_parser("reset", description="reset package settings - %s" % lic,
                                          parents=[shared_parser])

    argcomplete.autocomplete(main_parser)
    args, other_args = main_parser.parse_known_args()
    settings.configure_logging()
    colorama.init()

    config = settings.DEFAULT_PATH
    if hasattr(args, "config"):
        if args.config:
            config = args.config

    if args.subcommand == "show":
        if not args.brief and not args.config:
            logging.info(settings.DEFAULT_SETTINGS_HELP)
            logging.info(SEP + "\n" + config + "\n" + SEP)
        show(config, colored=args.no_color)
        if config == settings.DEFAULT_PATH and not args.brief:
            logging.info(SEP + "\nsee text above for parameter descriptions")

    elif args.subcommand == "set":
        if not os.access(config, os.W_OK):
            logging.info("no permission to modify " + config)
            sys.exit()
        if other_args or args.merge:
            logging.info(SEP + "\nold configuration:\n" + SEP)
            show(config, colored=args.no_color)
            try:
                set_cfg(config, other_args)
            except ConfigError as e:
                logging.error(e)
                sys.exit(1)
            if args.merge:
                merge_json_union(config, args.merge)
            logging.info(SEP + "\nnew configuration:\n" + SEP)
            show(config, colored=args.no_color)
        else:
            logging.info("no configuration parameters given (see --help)")

    elif args.subcommand == "generate":
        if other_args:
            logging.info(SEP + "\nparsed by argparse:\n" + str(other_args))
            logging.info(SEP + "\nWARNING:\n" +
                  "make sure you use the 'long-style' -- options (e.g. --plot) if possible\n" +
                  "and no combined 'short' - flags, (e.g. -avp)\n" + SEP)
            data = generate(other_args)
            log_info_dict_json(data, colored=args.no_color)
            if args.out and user.check_and_confirm_overwrite(args.out):
                with open(args.out, 'w') as out:
                    out.write(json.dumps(data, indent=4, sort_keys=True))
            elif not args.out:
                logging.info(SEP + "\nno output file specified (-o / --out) - doing nothing\n" + SEP)
        else:
            logging.info("no command line arguments given (see --help)")

    elif args.subcommand == "reset":
        if not os.access(config, os.W_OK):
            logging.info("no permission to modify" + config)
            sys.exit()
        if user.confirm("reset the package settings to the default settings? (y/n)"):
            reset_pkg_settings(settings.DEFAULT_PATH)
            logging.info(SEP + "\npackage settings after reset:\n" + SEP)
            show(settings.DEFAULT_PATH, colored=args.no_color)


if __name__ == '__main__':
    main()



# -*- coding: UTF8 -*-
"""
this module provides metrics for the evaluation of SLAM algorithms
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function  # Python 2.7 backwards compatibility

import abc
import logging
import math
import sys
from enum import Enum  # requires enum34 in Python 2.7

import numpy as np

from evo.algorithms import filters
from evo.algorithms import transformations as tr
from evo.algorithms import lie_algebra as lie

if sys.version_info[0] >= 3 and sys.version_info[1] >= 4:
    ABC = abc.ABC
else:
    ABC = abc.ABCMeta('ABC', (), {})


class MetricsException(Exception):
    pass


class StatisticsType(Enum):
    rmse = "rmse"
    mean = "mean"
    median = "median"
    std = "std"
    min = "min"
    max = "max"
    sse = "sse"


class PoseRelation(Enum):
    full_transformation = "full transformation"
    translation_part = "translation part"
    rotation_part = "rotation part"
    rotation_angle_rad = "rotation angle in radians"
    rotation_angle_deg = "rotation angle in degrees"


class Unit(Enum):
    meters = "m"
    seconds = "s"
    degrees = "deg"
    radians = "rad"
    frames = "frames"


class VelUnit(Enum):
    meters_per_sec = "m/s"
    rad_per_sec = "rad/s"
    degrees_per_sec = "deg/s"


class StateModel(Enum):
    """
    conventions for state vectors
    """
    xyz = 1,  # translation only
    xyz_rpy = 2,  # translation + roll, pitch, yaw
    xyz_quat_xyzw = 3,  # translation + xyzw quaternion
    xyz_quat_wxyz = 4,  # translation + wxyz quaternion


class Metric(ABC):
    @abc.abstractmethod
    def reset_parameters(self, parameters):
        return

    @abc.abstractmethod
    def process_data(self, data):
        return

    @abc.abstractmethod
    def get_statistic(self):
        return

    @abc.abstractmethod
    def get_all_statistics(self):
        return


class Hausdorff(Metric):
    def __init__(self, metric=lambda a, b: a - b, backward=True, two_way=False):
        self.metric = metric
        self.backward = backward
        self.two_way = two_way

    def reset_parameters(self, metric=lambda a, b: a - b, backward=True, two_way=False):
        self.__init__(metric, backward, two_way)

    def hausdorff_base(self, A, B):
        h = 0.0
        i = 0
        u = 0
        for a_i in A:
            print("\r" + str(i), end="")
            min_d = min([self.metric(a_i, b_i) for b_i in B])
            if min_d > h:
                h = min_d
                u = i
            i += 1
        print(" winner: " + str(u))
        return h

    def process_data(self, data):
        A, B = data
        if self.backward:
            return self.hausdorff_base(B, A)
        elif self.two_way:
            return max(self.hausdorff_base(A, B), self.hausdorff_base(B, A))
        else:
            return self.hausdorff_base(A, B)

    def get_statistic(self):
        pass

    def get_all_statistics(self):
        pass


class NEES(Metric):
    """
    NEES: normalized estimation error squared
    metric for determining the consistency of a state estimator w.r.t. a reference model
    """

    def __init__(self, threshold=0, state_model=StateModel.xyz_quat_wxyz):
        self.threshold = threshold
        self.state_model = state_model
        self.nees = []
        self.num_outliers = 0

    def reset_parameters(self, threshold, state_model=StateModel.xyz_quat_wxyz):
        """
        resets the current parameters and results
        :param threshold: values above will be counted as outliers
        :param state_model: StateModel value describing how the state vectors are formed
        """
        self.threshold = threshold
        self.state_model = state_model
        self.nees = []
        self.num_outliers = 0

    @staticmethod
    def nees_base(r, cov, threshold):
        """
        calculates the normalized estimation error squared (NEES) given a residual state vector
        :param r: residual state vector (estimate minus reference)
        :param cov: covariance matrix (inverse information matrix)
        :param threshold: values above will be marked as outliers
        :return: epsilon (NEES value), outlier (True or False)
        """
        n = np.dot(r.T, np.dot(np.linalg.inv(cov), r)),  # NEES value
        if np.abs(n)[0] > threshold:
            return n, True  # mark as outlier
        else:
            return n, False

    # TODO quaternion version not yet tested
    def calculate_residuals(self, states_ref, states_est):
        """
        calculate residual states ("inverse motion operator")
        :param states_ref: list of reference state vectors
        :param states_est: list of estimated state vectors
        """
        if self.state_model in {StateModel.xyz, StateModel.xyz_rpy}:
            residuals = states_ref - states_est
        elif self.state_model in {StateModel.xyz_quat_wxyz, StateModel.xyz_quat_xyzw}:
            if self.state_model == StateModel.xyz_quat_xyzw:
                # change quaternion order to wxyz for transformations.py
                states_ref[:, 3:] = np.roll(states_ref[:, 3:], -1, axis=1)
                states_est[:, 3:] = np.roll(states_est[:, 3:], -1, axis=1)
            residuals_xyz = states_ref[:, :3] - states_est[:, :3]
            residuals_quat = np.array(
                [tr.quaternion_multiply(q_1, tr.quaternion_inverse(q_2))  # quaternion division
                 for q_1, q_2 in zip(states_ref[:, 3:], states_est[:, 3:])])
            residuals = np.hstack((residuals_xyz, residuals_quat))
        else:
            raise MetricsException("unsupported state model")
        return residuals

    # TODO quaternion version not yet tested
    def process_data(self, data):
        """
        calculate the normalized estimation error squared (NEES)
        for a batch of state vectors and covariances
        :param data: tuple (states_ref, states_est, covariances) with:
        states_ref: list of reference state vectors
        states_est: list of estimated state vectors
        covariances: list of covariance matrices
        """
        if len(data) != 3:
            raise MetricsException(
                "please provide data tuple as: (states_ref, states_est, covariances)")
        states_ref, states_est, covariances = data
        if len(states_est) != len(states_ref) != len(covariances):
            raise MetricsException("data lists must have same length")
        # calculate NEES
        residuals = self.calculate_residuals(states_ref, states_est)
        for r, cov in zip(residuals, covariances):
            try:
                epsilon, is_outlier = self.nees_base(r, cov, self.threshold)
                self.nees.append(epsilon)
                if is_outlier:
                    self.num_outliers += 1
            except np.linalg.LinAlgError as e:
                logging.warning(str(e) + ", ignoring value for NEES calculation")

    def get_statistic(self, statistics_type=StatisticsType.mean):
        if statistics_type == StatisticsType.rmse:
            return np.sqrt(np.mean(np.power(self.nees, 2)))
        elif statistics_type == StatisticsType.mean:
            return np.mean(self.nees)
        elif statistics_type == StatisticsType.median:
            return np.median(self.nees)
        elif statistics_type == StatisticsType.std:
            return np.std(self.nees)
        elif statistics_type == StatisticsType.min:
            return np.min(np.absolute(self.nees))
        elif statistics_type == StatisticsType.max:
            return np.max(np.absolute(self.nees))
        else:
            raise MetricsException("unsupported statistics_type")

    def get_all_statistics(self):
        """
        shortcut for calling get_statistics with all supported StatisticTypes
        :return: a dictionary {StatisticsType_string : float}
        """
        statistics = {}
        for s in StatisticsType:
            try:
                statistics[s.value] = self.get_statistic(s)
            except MetricsException as e:
                if "unsupported statistics_type" not in str(e):  # ignore unsupported statistics
                    raise
        return statistics


class RPE(Metric):
    """
    RPE: relative pose error
    metric for investigating the odometry drift
    """

    def __init__(self, pose_relation=PoseRelation.translation_part, delta=1.0,
                 delta_unit=Unit.frames,
                 rel_delta_tol=0.1, all_pairs=False):
        if delta < 0:
            raise MetricsException("delta must be a positive number")
        if delta_unit == Unit.frames and not isinstance(delta, int) and not delta.is_integer():
            raise MetricsException(
                "delta must be integer (no decimals) for delta unit " + str(delta_unit))
        self.delta = int(delta) if delta_unit == Unit.frames else delta
        self.delta_unit = delta_unit
        self.rel_delta_tol = rel_delta_tol
        self.pose_relation = pose_relation
        self.all_pairs = all_pairs
        self.E = []
        self.error = []
        self.delta_ids = []
        if pose_relation == PoseRelation.translation_part:
            self.unit = Unit.meters
        elif pose_relation == PoseRelation.rotation_angle_deg:
            self.unit = Unit.degrees
        elif pose_relation == PoseRelation.rotation_angle_rad:
            self.unit = Unit.radians
        else:
            self.unit = None  # dimension-less

    def __str__(self):
        title = "RPE w.r.t. "
        title += (str(self.pose_relation.value) + " "
                  + ("(" + self.unit.value + ")" if self.unit else ""))
        title += "\nfor delta = " + str(self.delta) + " (" + str(self.delta_unit.value) + ")"
        if not self.all_pairs:
            title += " using consecutive pairs"
        else:
            title += " using all possible pairs"
        return title

    def reset_parameters(self, pose_relation=PoseRelation.translation_part, delta=1.0,
                         delta_unit=Unit.frames,
                         all_pairs=False):
        """
        resets the current parameters and results
        :param delta: the interval step for indices (default: 1)
        :param delta_unit: unit of delta (Unit enum member)
        :param pose_relation: MotionType value defining how the RPE should be calculated
        :param all_pairs: use all possible pairs instead of consecutive pairs
        """
        self.__init__(pose_relation, delta, delta_unit, all_pairs)

    @staticmethod
    def rpe_base(Q_i, Q_i_delta, P_i, P_i_delta):
        """
        relative SE(3) error pose for a single pose pair
        following the notation of the TUM RGB-D paper
        :param Q_i: reference SE(3) pose at i
        :param Q_i_delta: reference SE(3) pose at i+delta
        :param P_i: estimated SE(3) pose at i
        :param P_i_delta: estimated SE(3) pose at i+delta
        :return: the RPE matrix E_i in SE(3)
        """
        Q_rel = lie.relative_se3(Q_i, Q_i_delta)
        P_rel = lie.relative_se3(P_i, P_i_delta)
        E_i = lie.relative_se3(Q_rel, P_rel)
        return E_i

    def process_data(self, data, id_pairs=None):
        """
        calculate relative poses on a batch of SE(3) poses
        :param data: tuple (traj_ref, traj_est) with:
        traj_ref: reference evo.trajectory.PosePath or derived
        traj_est: estimated evo.trajectory.PosePath or derived
        :param id_pairs: pre-computed pair indices if you know what you're doing (ignores delta)
        """
        if len(data) != 2:
            raise MetricsException("please provide data tuple as: (traj_ref, traj_est)")
        traj_ref, traj_est = data
        if traj_ref.num_poses != traj_est.num_poses:
            raise MetricsException("trajectories must have same number of poses")

        if id_pairs is None:
            id_pairs = id_pairs_from_delta(traj_est.poses_se3, self.delta, self.delta_unit,
                                           self.rel_delta_tol, all_pairs=self.all_pairs)
        if not self.all_pairs:
            self.delta_ids = [j for i, j in id_pairs]  # store flat id list e.g. for plotting
        self.E = [self.rpe_base(traj_ref.poses_se3[i], traj_ref.poses_se3[j],
                                traj_est.poses_se3[i], traj_est.poses_se3[j])
                  for i, j in id_pairs]
        logging.debug("compared " + str(len(self.E)) + " relative pose pairs, delta = "
                      + str(self.delta) + " (" + str(self.delta_unit.value) + ") "
                      + ("with all possible pairs" if self.all_pairs else "with consecutive pairs"))

        logging.debug("calculating RPE for " + str(self.pose_relation.value) + " pose relation...")
        if self.pose_relation == PoseRelation.translation_part:
            self.error = [np.linalg.norm(E_i[:3, 3]) for E_i in self.E]
        elif self.pose_relation == PoseRelation.rotation_part:
            # ideal: rot(E_i) = 3x3 identity
            self.error = np.array(
                [np.linalg.norm(lie.so3_from_se3(E_i) - np.eye(3)) for E_i in self.E])
        elif self.pose_relation == PoseRelation.full_transformation:
            # ideal: E_i = 4x4 identity
            self.error = np.array([np.linalg.norm(E_i - np.eye(4)) for E_i in self.E])
        elif self.pose_relation == PoseRelation.rotation_angle_rad:
            self.error = np.array([abs(lie.so3_log(E_i[:3, :3])) for E_i in self.E])
        elif self.pose_relation == PoseRelation.rotation_angle_deg:
            self.error = np.array([abs(lie.so3_log(E_i[:3, :3])) * 180 / np.pi for E_i in self.E])
        else:
            raise MetricsException("unsupported pose_relation: ", self.pose_relation)

    def get_statistic(self, statistics_type=StatisticsType.rmse):
        """
        here, the statistics are the actual RPE value
        :param statistics_type: StatisticsType value indicating the type of averaging
        :return: a float RPE value
        """
        if statistics_type == StatisticsType.rmse:
            squared_errors = np.power(self.error, 2)
            return math.sqrt(np.mean(squared_errors))
        elif statistics_type == StatisticsType.sse:
            squared_errors = np.power(self.error, 2)
            return np.sum(squared_errors)
        elif statistics_type == StatisticsType.mean:
            return np.mean(self.error)
        elif statistics_type == StatisticsType.median:
            return np.median(self.error)
        elif statistics_type == StatisticsType.max:
            return np.max(self.error)
        elif statistics_type == StatisticsType.min:
            return np.min(self.error)
        elif statistics_type == StatisticsType.std:
            return np.std(self.error)
        else:
            raise MetricsException("unsupported statistics_type")

    def get_all_statistics(self):
        """
        shortcut for calling get_statistics with all supported StatisticTypes
        :return: a dictionary {StatisticsType_string : float}
        """
        statistics = {}
        for s in StatisticsType:
            try:
                statistics[s.value] = self.get_statistic(s)
            except MetricsException as e:
                if "unsupported statistics_type" not in str(e):  # ignore unsupported statistics
                    raise
        return statistics


class APE(Metric):
    """
    APE: absolute pose error
    metric for investigating the global consistency of a SLAM trajectory
    """

    def __init__(self, pose_relation=PoseRelation.translation_part):
        self.pose_relation = pose_relation
        self.E = []
        self.error = []
        if pose_relation == PoseRelation.translation_part:
            self.unit = Unit.meters
        elif pose_relation == PoseRelation.rotation_angle_deg:
            self.unit = Unit.degrees
        elif pose_relation == PoseRelation.rotation_angle_rad:
            self.unit = Unit.radians
        else:
            self.unit = None  # dimension-less

    def __str__(self):
        title = "APE w.r.t. "
        title += (str(self.pose_relation.value) + " "
                  + ("(" + self.unit.value + ")" if self.unit else ""))
        return title

    def reset_parameters(self, pose_relation=PoseRelation.translation_part):
        """
        resets the current parameters and results
        :param pose_relation: PoseRelation value defining how the RPE should be calculated
        """
        self.__init__(pose_relation)

    @staticmethod
    def ape_base(x_t, x_t_star):
        """
        absolute error pose for a single SE(3) pose pair
        following the notation of the Kümmerle paper
        :param x_t: estimated absolute pose at t
        :param x_t_star: reference absolute pose at t
        .:return: the delta pose
        """
        return lie.relative_se3(x_t, x_t_star)

    def process_data(self, data):
        """
        calculate APE metric on a batch of poses
        :param data: tuple (traj_ref, traj_est) with:
        traj_ref: reference evo.trajectory.PosePath or derived
        traj_est: estimated evo.trajectory.PosePath or derived
        """
        if len(data) != 2:
            raise MetricsException("please provide data tuple as: (traj_ref, traj_est)")
        traj_ref, traj_est = data
        if traj_ref.num_poses != traj_est.num_poses:
            raise MetricsException("trajectories must have same number of poses")

        if self.pose_relation == PoseRelation.translation_part:
            # don't require full SE(3) matrices for faster computation
            self.E = traj_est.positions_xyz - traj_ref.positions_xyz
        else:
            self.E = [self.ape_base(x_t, x_t_star) for x_t, x_t_star in zip(traj_est.poses_se3, traj_ref.poses_se3)]
        logging.debug("compared " + str(len(self.E)) + " absolute pose pairs")
        logging.debug("calculating APE for " + str(self.pose_relation.value) + " pose relation...")
        if self.pose_relation == PoseRelation.translation_part:
            # E is an array of position vectors only in this case
            self.error = [np.linalg.norm(E_i) for E_i in self.E]
        elif self.pose_relation == PoseRelation.rotation_part:
            self.error = np.array(
                [np.linalg.norm(lie.so3_from_se3(E_i) - np.eye(3)) for E_i in self.E])
        elif self.pose_relation == PoseRelation.full_transformation:
            self.error = np.array([np.linalg.norm(E_i - np.eye(4)) for E_i in self.E])
        elif self.pose_relation == PoseRelation.rotation_angle_rad:
            self.error = np.array([abs(lie.so3_log(E_i[:3, :3])) for E_i in self.E])
        elif self.pose_relation == PoseRelation.rotation_angle_deg:
            self.error = np.array([abs(lie.so3_log(E_i[:3, :3])) * 180 / np.pi for E_i in self.E])
        else:
            raise MetricsException("unsupported pose_relation")

    def get_statistic(self, statistics_type=StatisticsType.rmse):
        """
        here, the statistics are the actual APE value
        :param statistics_type: StatisticsType value indicating the type of averaging
        :return: a float APE value
        """
        if statistics_type == StatisticsType.rmse:
            squared_errors = np.power(self.error, 2)
            return math.sqrt(np.mean(squared_errors))
        elif statistics_type == StatisticsType.sse:
            squared_errors = np.power(self.error, 2)
            return np.sum(squared_errors)
        elif statistics_type == StatisticsType.mean:
            return np.mean(self.error)
        elif statistics_type == StatisticsType.median:
            return np.median(self.error)
        elif statistics_type == StatisticsType.max:
            return np.max(self.error)
        elif statistics_type == StatisticsType.min:
            return np.min(self.error)
        elif statistics_type == StatisticsType.std:
            return np.std(self.error)
        else:
            raise MetricsException("unsupported statistics_type")

    def get_all_statistics(self):
        """
        shortcut for calling get_statistics with all supported StatisticTypes
        :return: a dictionary {StatisticsType_string : float}
        """
        statistics = {}
        for s in StatisticsType:
            try:
                statistics[s.value] = self.get_statistic(s)
            except MetricsException as e:
                if "unsupported statistics_type" not in str(e):  # ignore unsupported statistics
                    raise
        return statistics


def id_pairs_from_delta(poses, delta, delta_unit, rel_tol=0.1, all_pairs=False):
    """
    get index tuples of pairs with distance==delta from a pose list
    :param poses: list of SE(3) poses
    :param delta: the interval step for indices
    :param delta_unit: unit of delta (Unit enum member)
    :param rel_tol: relative tolerance to accept or reject deltas
    :param all_pairs: use all possible pairs instead of consecutive pairs
    :return: list of index tuples (pairs)
    """
    if delta_unit == Unit.frames:
        id_pairs = filters.filter_pairs_by_index(poses, delta, all_pairs)
    elif delta_unit == Unit.meters:
        id_pairs = filters.filter_pairs_by_path(poses, delta, delta * rel_tol, all_pairs)
    elif delta_unit in {Unit.degrees, Unit.radians}:
        use_degrees = (delta_unit == Unit.degrees)
        id_pairs = filters.filter_pairs_by_angle(poses, delta, delta * rel_tol, use_degrees,
                                                 all_pairs)
    else:
        raise MetricsException("unsupported delta unit: " + str(delta_unit))

    if len(id_pairs) == 0:
        raise MetricsException("delta = " + str(delta) + " (" + str(delta_unit.value) + ") " +
                               "produced empty index list - try lower values or higher tolerance")
    logging.debug("found " + str(len(id_pairs)) + " pairs with delta " + str(delta)
                  + " (" + str(delta_unit.value) + ") among " + str(len(poses)) + " poses "
                  + ("using consecutive pairs " if not all_pairs else "using all possible pairs"))
    return id_pairs



from setuptools import setup
from setuptools.command.install import install

import os
import sys
import subprocess as sp

# monkey patch because setuptools entry_points are slow as fuck
# https://github.com/ninjaaron/fast-entry_points
import fastentrypoints


def python_below_34():
    return sys.version_info[0] < 3 or sys.version_info[1] < 4


def _post_install(install_lib_dir):
    print("activating argcomplete")
    sp.call("activate-global-python-argcomplete", shell=True)
    print("done - argcomplete should work now")
    # sp.call(["jupyter", "nbextension", "enable", "--py", "--sys-prefix", "widgetsnbextension"])


# https://stackoverflow.com/a/18159969/6288017
class CustomInstall(install):
    def run(self):
        install.run(self)
        self.execute(_post_install, (self.install_lib,), msg="running post install task")


setup(name="evo",
      version=open("evo/version").read(),
      description="Python package for the evaluation of odometry and SLAM",
      author="Michael Grupp",
      author_email="michael.grupp@tum.de",
      url="https://github.com/MichaelGrupp/evo",
      license="GNU GPL v3",
      keywords=["SLAM", "odometry", "vision", "laser", "visual", "robotics", "evaluation", "metric"],
      packages=["evo", "evo.algorithms", "evo.tools"],
      package_data={"evo": ["README.md", "settings.json", "version", "LICENSE"]},
      entry_points={"console_scripts": ["evo_ape=evo.entry_points:ape",
                                        "evo_rpe=evo.entry_points:rpe",
                                        "evo_traj=evo.entry_points:traj",
                                        "evo_res=evo.entry_points:res",
                                        "evo_rpe-for-each=evo.entry_points:rpe_for_each",
                                        "evo_config=evo.main_config:main",
                                        "evo_fig=evo.main_fig:main",
                                        "evo=evo.main_evo:main"]},
      zip_safe=False,
      install_requires=["numpy", "matplotlib", "scipy", "pandas", "seaborn",
                        "natsort", "argcomplete", "colorama", "pygments"]  # "jupyter"
                       + (["enum34"] if python_below_34() else []),
      cmdclass={"install": CustomInstall}
      )




#!/usr/bin/env python
# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
main executable for calculating the relative pose error (RPE) metric
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import logging

SEP = "-" * 80  # separator line


def parser():
    import argparse
    basic_desc = "relative pose error (RPE) metric app"
    lic = "(c) michael.grupp@tum.de"
    shared_parser = argparse.ArgumentParser(add_help=False)
    algo_opts = shared_parser.add_argument_group("algorithm options")
    output_opts = shared_parser.add_argument_group("output options")
    usability_opts = shared_parser.add_argument_group("usability options")
    algo_opts.add_argument("-r", "--pose_relation",
                           help="pose relation on which the RPE is based", default="trans_part",
                           choices=["full", "trans_part", "rot_part", "angle_deg", "angle_rad"])
    algo_opts.add_argument("-d", "--delta", help="delta between relative poses",
                           default=1, type=float)
    algo_opts.add_argument("-t", "--delta_tol", help="relative delta tolerance for all_pairs mode",
                           default=0.1, type=float)
    algo_opts.add_argument("-u", "--delta_unit",
                           help="unit of delta - `f` (frames), `d` (deg), `r` (rad), `m`(meters)",
                           default="f", choices=['f', 'd', 'r', 'm'])
    algo_opts.add_argument("--all_pairs", help="use all possible pairs (incompatible with plot)",
                           action="store_true")
    algo_opts.add_argument("-a", "--align", help="alignment with Umeyama's method (no scale) - "
                                                 "only useful for plotting with RPE",
                           action="store_true")
    algo_opts.add_argument("-s", "--correct_scale", help="correct scale with Umeyama's method",
                           action="store_true")
    output_opts.add_argument("-p", "--plot", help="show plot window", action="store_true")
    output_opts.add_argument("--plot_mode", help="the axes for  plot projection",
                             default=None, choices=["xy", "yx", "xz", "zx", "yz", "xyz"])
    output_opts.add_argument("--save_plot", help="path to save plot", default=None)
    output_opts.add_argument("--serialize_plot", help="path to serialize plot (experimental)",
                             default=None)
    output_opts.add_argument("--save_results", help=".zip file path to store results")
    usability_opts.add_argument("--no_warnings", help="no warnings requiring user confirmation",
                                action="store_true")
    usability_opts.add_argument("-v", "--verbose", help="verbose output", action="store_true")
    usability_opts.add_argument("--silent", help="don't print any output", action="store_true")
    usability_opts.add_argument("--debug", help="verbose output with additional debug info",
                                action="store_true")
    usability_opts.add_argument("-c", "--config",
                                help=".json file with parameters (priority over command line args)")

    main_parser = argparse.ArgumentParser(description="%s %s" % (basic_desc, lic))
    sub_parsers = main_parser.add_subparsers(dest="subcommand")
    sub_parsers.required = True
    kitti_parser = sub_parsers.add_parser("kitti", description="%s for KITTI pose files - %s"
                                                               % (basic_desc, lic),
                                          parents=[shared_parser])
    kitti_parser.add_argument("ref_file", help="reference pose file (ground truth)")
    kitti_parser.add_argument("est_file", help="estimated pose file")

    tum_parser = sub_parsers.add_parser("tum", description="%s for TUM trajectory files - %s"
                                                           % (basic_desc, lic),
                                        parents=[shared_parser])
    tum_parser.add_argument("ref_file", help="reference trajectory file")
    tum_parser.add_argument("est_file", help="estimated trajectory file")
    tum_parser.add_argument("--t_max_diff",
                            help="maximum timestamp difference for data association",
                            default=0.01, type=float)
    tum_parser.add_argument("--t_offset", help="constant timestamp offset for data association",
                            default=0.0, type=float)

    euroc_parser = sub_parsers.add_parser("euroc", description="%s for EuRoC MAV .csv's - %s"
                                                               % (basic_desc, lic),
                                          parents=[shared_parser])
    euroc_parser.add_argument("state_gt_csv",
                              help="ground truth: <seq>/mav0/state_groundtruth_estimate0/data.csv")
    euroc_parser.add_argument("est_file", help="estimated trajectory file in TUM format")
    euroc_parser.add_argument("--t_max_diff",
                              help="maximum timestamp difference for data association",
                              default=0.01, type=float)
    euroc_parser.add_argument("--t_offset", help="constant timestamp offset for data association",
                              default=0.0, type=float)

    bag_parser = sub_parsers.add_parser("bag", description="%s for ROS bag files - %s"
                                                           % (basic_desc, lic),
                                        parents=[shared_parser])
    bag_parser.add_argument("bag", help="ROS bag file")
    bag_parser.add_argument("ref_topic", help="reference geometry_msgs/PoseStamped topic")
    bag_parser.add_argument("est_topic", help="estimated geometry_msgs/PoseStamped topic")
    bag_parser.add_argument("--t_max_diff",
                            help="maximum timestamp difference for data association",
                            default=0.01, type=float)
    bag_parser.add_argument("--t_offset", help="constant timestamp offset for data association",
                            default=0.0, type=float)
    return main_parser


def main_rpe(traj_ref, traj_est, pose_relation, delta, delta_unit,
             rel_delta_tol=0.1, all_pairs=False, align=False, correct_scale=False,
             ref_name="", est_name="", show_plot=False, save_plot=None,
             plot_mode=None, save_results=None, no_warnings=False, support_loop=False,
             serialize_plot=None):

    from evo.algorithms import metrics
    from evo.algorithms import trajectory
    from evo.tools import file_interface
    from evo.tools.settings import SETTINGS

    if (show_plot or save_plot or serialize_plot) and all_pairs:
        raise metrics.MetricsException("all_pairs mode cannot be used with plotting functions")

    only_scale = correct_scale and not align
    if align or correct_scale:
        logging.debug(SEP)
        if only_scale:
            logging.debug("correcting scale...")
        else:
            logging.debug("aligning using Umeyama's method..."
                          + (" (with scale correction)" if correct_scale else ""))
        traj_est = trajectory.align_trajectory(traj_est, traj_ref, correct_scale, only_scale)
    logging.debug(SEP)

    # calculate RPE
    data = (traj_ref, traj_est)
    rpe_metric = metrics.RPE(pose_relation, delta, delta_unit, rel_delta_tol, all_pairs)
    rpe_metric.process_data(data)
    rpe_statistics = rpe_metric.get_all_statistics()

    title = str(rpe_metric)
    if align and not correct_scale:
        title += "\n(with SE(3) Umeyama alignment)"
    elif align and correct_scale:
        title += "\n(with Sim(3) Umeyama alignment)"
    elif only_scale:
        title += "\n(scale corrected)"
    else:
        title += "\n(not aligned)"
    logging.debug(SEP)
    res_str = ""
    for name, val in sorted(rpe_statistics.items()):
        res_str += "{:>10}".format(name) + "\t" + "{0:.6f}".format(val) + "\n"
    logging.info("\nstatistics of " + title + ":\n\n" + res_str)

    seconds_from_start = None
    if show_plot or save_plot or serialize_plot or save_results and not all_pairs:
        # restrict trajectories to delta ids for plot
        if support_loop:
            # avoid overwriting if called repeatedly e.g. in Jupyter notebook
            import copy
            traj_ref = copy.deepcopy(traj_ref)
            traj_est = copy.deepcopy(traj_est)
        traj_ref.reduce_to_ids(rpe_metric.delta_ids)
        traj_est.reduce_to_ids(rpe_metric.delta_ids)
        if isinstance(traj_est, trajectory.PoseTrajectory3D):
            seconds_from_start = [t - traj_est.timestamps[0] for t in traj_est.timestamps]
        else:
            seconds_from_start = None

        if show_plot or save_plot or serialize_plot:
            from evo.tools import plot
            import matplotlib.pyplot as plt
            logging.debug(SEP)
            logging.debug("plotting results... ")
            fig1 = plt.figure(figsize=(SETTINGS.plot_figsize[0], SETTINGS.plot_figsize[1]))
            # metric values
            plot.error_array(fig1, rpe_metric.error, x_array=seconds_from_start,
                             statistics=rpe_statistics,
                             name="RPE" + (
                             " (" + rpe_metric.unit.value + ")") if rpe_metric.unit else "",
                             title=title, xlabel="$t$ (s)" if seconds_from_start else "index")
            # info text
            if SETTINGS.plot_info_text and est_name and ref_name:
                ax = fig1.gca()
                ax.text(0, -0.12, "estimate:  " + est_name + "\nreference: " + ref_name,
                        transform=ax.transAxes, fontsize=8, color="gray")
            # trajectory colormapped
            fig2 = plt.figure(figsize=(SETTINGS.plot_figsize[0], SETTINGS.plot_figsize[1]))
            plot_mode = plot_mode if plot_mode is not None else plot.PlotMode.xyz
            ax = plot.prepare_axis(fig2, plot_mode)
            plot.traj(ax, plot_mode, traj_ref, '--', 'gray', 'reference',
                      alpha=0 if SETTINGS.plot_hideref else 1)
            plot.traj_colormap(ax, traj_est, rpe_metric.error, plot_mode,
                               min_map=rpe_statistics["min"], max_map=rpe_statistics["max"],
                               title="RPE mapped onto trajectory")
            fig2.axes.append(ax)
            plot_collection = plot.PlotCollection(title)
            plot_collection.add_figure("raw", fig1)
            plot_collection.add_figure("map", fig2)
            if show_plot:
                plot_collection.show()
            if save_plot:
                plot_collection.export(save_plot, confirm_overwrite=not no_warnings)
            if serialize_plot:
                logging.debug(SEP)
                plot_collection.serialize(serialize_plot, confirm_overwrite=not no_warnings)

    if save_results:
        logging.debug(SEP)
        file_interface.save_res_file(save_results, rpe_metric, rpe_statistics,
                                        title, ref_name, est_name, seconds_from_start, traj_ref,
                                        traj_est, confirm_overwrite=not no_warnings)

    return rpe_statistics, rpe_metric.error


def run(args):
    from evo.algorithms import metrics
    from evo.tools import file_interface, settings

    settings.configure_logging(args.verbose, args.silent, args.debug)
    if args.debug:
        import pprint
        logging.debug("main_parser config:\n"
                      + pprint.pformat({arg: getattr(args, arg) for arg in vars(args)}) + "\n")
    logging.debug(SEP)

    pose_relation = None
    if args.pose_relation == "full":
        pose_relation = metrics.PoseRelation.full_transformation
    elif args.pose_relation == "rot_part":
        pose_relation = metrics.PoseRelation.rotation_part
    elif args.pose_relation == "trans_part":
        pose_relation = metrics.PoseRelation.translation_part
    elif args.pose_relation == "angle_deg":
        pose_relation = metrics.PoseRelation.rotation_angle_deg
    elif args.pose_relation == "angle_rad":
        pose_relation = metrics.PoseRelation.rotation_angle_rad

    delta_unit = None
    if args.delta_unit == "f":
        delta_unit = metrics.Unit.frames
    elif args.delta_unit == "d":
        delta_unit = metrics.Unit.degrees
    elif args.delta_unit == "r":
        delta_unit = metrics.Unit.radians
    elif args.delta_unit == "m":
        delta_unit = metrics.Unit.meters

    traj_ref, traj_est, stamps_est = None, None, None
    ref_name, est_name = "", ""
    plot_mode = None  # no plot imports unless really needed (slow)
    if args.subcommand == "tum":
        traj_ref, traj_est = file_interface.load_assoc_tum_trajectories(
            args.ref_file,
            args.est_file,
            args.t_max_diff,
            args.t_offset,
        )
        ref_name, est_name = args.ref_file, args.est_file
        if args.plot or args.save_plot:
            from evo.tools.plot import PlotMode
            plot_mode = PlotMode.xyz if not args.plot_mode else PlotMode[args.plot_mode]
    elif args.subcommand == "kitti":
        traj_ref = file_interface.read_kitti_poses_file(args.ref_file)
        traj_est = file_interface.read_kitti_poses_file(args.est_file)
        ref_name, est_name = args.ref_file, args.est_file
        if args.plot or args.save_plot:
            from evo.tools.plot import PlotMode
            plot_mode = PlotMode.xz if not args.plot_mode else PlotMode[args.plot_mode]
    elif args.subcommand == "euroc":
        args.align = True
        logging.info("forcing trajectory alignment implicitly (EuRoC ground truth is in IMU frame)")
        logging.debug(SEP)
        traj_ref, traj_est = file_interface.load_assoc_euroc_trajectories(
            args.state_gt_csv,
            args.est_file,
            args.t_max_diff,
            args.t_offset,
        )
        ref_name, est_name = args.state_gt_csv, args.est_file
        if args.plot or args.save_plot:
            from evo.tools.plot import PlotMode
            plot_mode = PlotMode.xyz if not args.plot_mode else PlotMode[args.plot_mode]
    elif args.subcommand == "bag":
        import rosbag
        logging.debug("opening bag file " + args.bag)
        bag = rosbag.Bag(args.bag, 'r')
        try:
            traj_ref, traj_est = file_interface.load_assoc_bag_trajectories(
                bag,
                args.ref_topic,
                args.est_topic,
                args.t_max_diff,
                args.t_offset,
            )
        finally:
            bag.close()
        ref_name, est_name = args.ref_topic, args.est_topic
        if args.plot or args.save_plot:
            from evo.tools.plot import PlotMode
            plot_mode = PlotMode.xy if not args.plot_mode else PlotMode[args.plot_mode]

    main_rpe(traj_ref, traj_est, pose_relation, args.delta, delta_unit, args.delta_tol,
             args.all_pairs, args.align, args.correct_scale,
             ref_name, est_name, args.plot, args.save_plot, plot_mode,
             args.save_results, args.no_warnings, serialize_plot=args.serialize_plot)


if __name__ == '__main__':
    from evo import entry_points
    entry_points.rpe()
